# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
ROI editor
==========

The :mod:`cdl.core.gui.roieditor` module provides the ROI editor widgets
for signals and images.

Signal ROI editor
-----------------

.. autoclass:: SignalROIEditor
    :members:

Image ROI editor
----------------

.. autoclass:: ImageROIEditor
    :members:
"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...

from __future__ import annotations

import abc
from typing import TYPE_CHECKING, Generic, TypeVar, Union

from guidata.configtools import get_icon
from guidata.qthelpers import add_actions, create_action
from plotpy.builder import make
from plotpy.interfaces import IImageItemType
from plotpy.items import (
    AnnotatedCircle,
    AnnotatedPolygon,
    AnnotatedRectangle,
    CurveItem,
    MaskedImageItem,
    ObjectInfo,
    XRangeSelection,
)
from plotpy.plot import PlotDialog, PlotManager
from plotpy.tools import CircleTool, HRangeTool, PolygonTool, RectangleTool
from qtpy import QtCore as QC
from qtpy import QtWidgets as QW

import cdl.obj as dlo
from cdl.config import Conf, _
from cdl.core.model.base import (
    TypeObj,
    TypePlotItem,
    TypeROI,
    TypeROIItem,
    configure_roi_item,
)
from cdl.core.model.image import CircularROI, PolygonalROI, RectangularROI
from cdl.core.model.signal import SegmentROI
from cdl.env import execenv
from cdl.obj import ImageObj, ImageROI, SignalObj, SignalROI

if TYPE_CHECKING:
    from plotpy.plot import BasePlot
    from plotpy.tools.base import InteractiveTool


def plot_item_to_single_roi(
    item: XRangeSelection | AnnotatedRectangle | AnnotatedCircle | AnnotatedPolygon,
) -> SegmentROI | CircularROI | RectangularROI | PolygonalROI:
    """Factory function to create a single ROI object from plot item

    Args:
        item: Plot item

    Returns:
        Single ROI object
    """
    if isinstance(item, XRangeSelection):
        cls = SegmentROI
    elif isinstance(item, AnnotatedRectangle):
        cls = RectangularROI
    elif isinstance(item, AnnotatedCircle):
        cls = CircularROI
    elif isinstance(item, AnnotatedPolygon):
        cls = PolygonalROI
    else:
        raise TypeError(f"Unsupported ROI item type: {type(item)}")
    return cls.from_plot_item(item)


ROI_EDITOR_TOOLBAR_ID = "roi_editor_toolbar"


def configure_roi_item_in_tool(shape, obj: dlo.SignalObj | dlo.ImageObj) -> None:
    """Configure ROI item in tool"""
    fmt = obj.get_metadata_option("format")
    configure_roi_item(shape, fmt, lbl=True, editable=True, option=obj.PREFIX)


def tool_activate(tool: InteractiveTool) -> None:
    """Tool activate"""
    plot = tool.get_active_plot()
    plot.select_some_items([])  # Deselect all items
    tool.activate()


def tool_setup_shape(shape: TypeROIItem, obj: dlo.SignalObj | dlo.ImageObj) -> None:
    """Tool setup shape"""
    shape.setTitle("ROI")
    configure_roi_item_in_tool(shape, obj)


class ROISegmentTool(HRangeTool):
    """ROI segment tool"""

    TITLE = _("Add ROI")
    ICON = "signal_roi.svg"

    def __init__(self, manager: PlotManager, obj: dlo.SignalObj) -> None:
        super().__init__(
            manager, switch_to_default_tool=True, toolbar_id=ROI_EDITOR_TOOLBAR_ID
        )
        self.roi = SegmentROI([0, 1], False)
        self.obj = obj
        self.activate = tool_activate

    def create_shape(self) -> XRangeSelection:
        """Create shape"""
        shape = self.roi.to_plot_item(self.obj)
        configure_roi_item_in_tool(shape, self.obj)
        shape.selected = True
        return shape


class ROIRectangleTool(RectangleTool):
    """ROI rectangle tool"""

    TITLE = _("Add rectangular ROI")
    ICON = "roi_new_rectangle.svg"

    def __init__(self, manager: PlotManager, obj: dlo.ImageObj) -> None:
        super().__init__(
            manager,
            switch_to_default_tool=True,
            toolbar_id=ROI_EDITOR_TOOLBAR_ID,
            setup_shape_cb=tool_setup_shape,
        )
        self.roi = RectangularROI([0, 0, 1, 1], True)
        self.obj = obj
        self.activate = tool_activate

    def create_shape(self) -> tuple[AnnotatedRectangle, int, int]:
        """Create shape"""
        item = self.roi.to_plot_item(self.obj)
        return item, 0, 2

    def setup_shape(self, shape: AnnotatedRectangle) -> None:
        """Setup shape"""
        tool_setup_shape(shape, self.obj)


class ROICircleTool(CircleTool):
    """ROI circle tool"""

    TITLE = _("Add circular ROI")
    ICON = "roi_new_circle.svg"

    def __init__(self, manager: PlotManager, obj: dlo.ImageObj) -> None:
        super().__init__(
            manager,
            switch_to_default_tool=True,
            toolbar_id=ROI_EDITOR_TOOLBAR_ID,
            setup_shape_cb=tool_setup_shape,
        )
        self.roi = CircularROI([0, 0, 1], True)
        self.obj = obj
        self.activate = tool_activate

    def create_shape(self) -> tuple[AnnotatedCircle, int, int]:
        """Create shape"""
        item = self.roi.to_plot_item(self.obj)
        return item, 0, 1

    def setup_shape(self, shape: AnnotatedCircle) -> None:
        """Setup shape"""
        tool_setup_shape(shape, self.obj)


class ROIPolygonTool(PolygonTool):
    """ROI polygon tool"""

    TITLE = _("Add polygonal ROI")
    ICON = "roi_new_polygon.svg"

    def __init__(self, manager: PlotManager, obj: dlo.ImageObj) -> None:
        super().__init__(
            manager,
            switch_to_default_tool=True,
            toolbar_id=ROI_EDITOR_TOOLBAR_ID,
            setup_shape_cb=tool_setup_shape,
        )
        self.roi = PolygonalROI([[0, 0], [1, 0], [1, 1], [0, 1]], True)
        self.obj = obj
        self.activate = tool_activate

    def create_shape(self) -> tuple[AnnotatedPolygon, int, int]:
        """Create shape"""
        return self.roi.to_plot_item(self.obj)

    def setup_shape(self, shape: AnnotatedPolygon) -> None:
        """Setup shape"""
        tool_setup_shape(shape, self.obj)


TypeROIEditor = TypeVar("TypeROIEditor", bound="BaseROIEditor")


class BaseROIEditorMeta(type(QW.QWidget), abc.ABCMeta):
    """Mixed metaclass to avoid conflicts"""


class BaseROIEditor(
    QW.QWidget,
    Generic[TypeObj, TypeROI, TypePlotItem, TypeROIItem],
    metaclass=BaseROIEditorMeta,
):
    """ROI Editor"""

    ICON_NAME = None
    OBJ_NAME = None
    ROI_ITEM_TYPES = ()

    def __init__(
        self,
        parent: PlotDialog,
        obj: TypeObj,
        extract: bool,
        item: TypePlotItem | None = None,
    ) -> None:
        super().__init__(parent)
        self.plot_dialog = parent
        parent.accepted.connect(self.dialog_accepted)
        self.plot = parent.get_plot()
        self.toolbar = QW.QToolBar(self)
        self.obj = obj
        self.extract = extract
        self.__modified: bool | None = None

        roi = obj.roi
        if roi is None:
            roi = self.get_obj_roi_class()()
        self.__roi: TypeROI = roi

        fmt = obj.get_metadata_option("format")
        self.roi_items: list[TypeROIItem] = list(
            self.__roi.iterate_roi_items(obj, fmt, True, True)
        )

        self.add_tools_to_plot_dialog()
        item = obj.make_item() if item is None else item
        item.set_selectable(False)
        item.set_readonly(True)
        self.plot.add_item(item)
        for roi_item in self.roi_items:
            self.plot.add_item(roi_item)
            self.plot.set_active_item(roi_item)

        self.remove_all_action: QW.QAction | None = None
        self.singleobj_btn: QW.QToolButton | None = None

        self.setup_widget()

        # force update of ROI titles and remove_all_btn state
        self.items_changed(self.plot)

        self.plot.SIG_ITEMS_CHANGED.connect(self.items_changed)
        self.plot.SIG_ITEM_REMOVED.connect(self.item_removed)
        self.plot.SIG_RANGE_CHANGED.connect(lambda _rng, _min, _max: self.item_moved())
        self.plot.SIG_ANNOTATION_CHANGED.connect(lambda _plt: self.item_moved())

        #  In "extract mode", the dialog box OK button should always been enabled
        #  when at least one ROI is defined,
        #  whereas in non-extract mode (when editing ROIs) the OK button is by default
        #  disabled (until ROI data is modified)
        if extract:
            self.modified = len(self.roi_items) > 0
        else:
            self.modified = False

    @abc.abstractmethod
    def get_obj_roi_class(self) -> type[TypeROI]:
        """Get object ROI class"""

    @abc.abstractmethod
    def add_tools_to_plot_dialog(self) -> None:
        """Add tools to plot dialog"""

    @property
    def modified(self) -> bool:
        """Return dialog modified state"""
        return self.__modified

    @modified.setter
    def modified(self, value: bool) -> None:
        """Set dialog modified state"""
        self.__modified = value
        if self.extract:
            #  In "extract mode", OK button is enabled when at least one ROI is defined
            value = value and len(self.roi_items) > 0
        self.plot_dialog.button_box.button(QW.QDialogButtonBox.Ok).setEnabled(value)

    def dialog_accepted(self) -> None:
        """Parent dialog was accepted: updating ROI Editor data"""
        self.__roi.empty()
        for roi_item in self.roi_items:
            self.__roi.add_roi(plot_item_to_single_roi(roi_item))
        if self.singleobj_btn is not None:
            singleobj = self.singleobj_btn.isChecked()
            self.__roi.singleobj = singleobj
            Conf.proc.extract_roi_singleobj.set(singleobj)

    def get_roieditor_results(self) -> tuple[TypeROI, bool]:
        """Get ROI editor results

        Returns:
            A tuple containing the ROI data parameters and ROI modified state.
            ROI modified state is True if the ROI data has been modified within
            the dialog box.
        """
        return self.__roi, self.modified

    def create_actions(self) -> list[QW.QAction]:
        """Create actions"""
        self.remove_all_action = create_action(
            self,
            _("Remove all ROIs"),
            icon=get_icon("roi_delete.svg"),
            triggered=self.remove_all_rois,
        )
        return [None, self.remove_all_action]

    def setup_widget(self) -> None:
        """Setup ROI editor widget"""
        layout = QW.QHBoxLayout()
        self.toolbar.setToolButtonStyle(QC.Qt.ToolButtonTextUnderIcon)
        add_actions(self.toolbar, self.create_actions())
        layout.addWidget(self.toolbar)
        if self.extract:
            self.singleobj_btn = QW.QCheckBox(
                _("Extract all ROIs into a single %s object") % self.OBJ_NAME,
                self,
            )
            layout.addWidget(self.singleobj_btn)
            self.singleobj_btn.setChecked(self.__roi.singleobj)
        layout.addStretch()
        self.setLayout(layout)

    def remove_all_rois(self) -> None:
        """Remove all ROIs"""
        if (
            execenv.unattended
            or QW.QMessageBox.question(
                self,
                _("Remove all ROIs"),
                _("Are you sure you want to remove all ROIs?"),
            )
            == QW.QMessageBox.Yes
        ):
            self.plot.del_items(self.roi_items)

    @abc.abstractmethod
    def update_roi_titles(self) -> None:
        """Update ROI annotation titles"""

    def update_roi_items(self) -> None:
        """Update ROI items"""
        old_nb_items = len(self.roi_items)
        self.roi_items = [
            item
            for item in self.plot.get_items()
            if isinstance(item, self.ROI_ITEM_TYPES)
        ]
        self.plot.select_some_items([])
        self.update_roi_titles()
        if old_nb_items != len(self.roi_items):
            self.modified = True

    def items_changed(self, _plot: BasePlot) -> None:
        """Items have changed"""
        self.update_roi_items()
        self.remove_all_action.setEnabled(len(self.roi_items) > 0)

    def item_removed(self, item) -> None:
        """Item was removed. Since all items are read-only except ROIs...
        this must be an ROI."""
        assert item in self.roi_items
        self.update_roi_items()
        self.modified = True

    def item_moved(self) -> None:
        """ROI plot item has just been moved"""
        self.modified = True


class ROIRangeInfo(ObjectInfo):
    """ObjectInfo for ROI selection"""

    def __init__(self, roi_items) -> None:
        self.roi_items = roi_items

    def get_text(self) -> str:
        textlist = []
        for index, roi_item in enumerate(self.roi_items):
            x0, x1 = roi_item.get_range()
            textlist.append(f"ROI{index:02d}: {x0} ≤ x ≤ {x1}")
        return "<br>".join(textlist)


class SignalROIEditor(BaseROIEditor[SignalObj, SignalROI, CurveItem, XRangeSelection]):
    """Signal ROI Editor"""

    ICON_NAME = "signal_roi.svg"
    OBJ_NAME = _("signal")
    ROI_ITEM_TYPES = (XRangeSelection,)

    def get_obj_roi_class(self) -> type[SignalROI]:
        """Get object ROI class"""
        return SignalROI

    def add_tools_to_plot_dialog(self) -> None:
        """Add tools to plot dialog"""
        mgr = self.plot_dialog.get_manager()
        mgr.add_toolbar(self.toolbar, ROI_EDITOR_TOOLBAR_ID)
        mgr.add_tool(ROISegmentTool, self.obj)

    def setup_widget(self) -> None:
        """Setup ROI editor widget"""
        super().setup_widget()
        info = ROIRangeInfo(self.roi_items)
        info_label = make.info_label("BL", info, title=_("Regions of interest"))
        self.plot.add_item(info_label)
        self.info_label = info_label

    def update_roi_titles(self):
        """Update ROI annotation titles"""
        super().update_roi_titles()
        self.info_label.update_text()


class ImageROIEditor(
    BaseROIEditor[
        ImageObj,
        ImageROI,
        MaskedImageItem,
        # `Union` is mandatory here for Python 3.9-3.10 compatibility:
        Union[AnnotatedPolygon, AnnotatedRectangle, AnnotatedCircle],
    ]
):
    """Image ROI Editor"""

    ICON_NAME = "image_roi.svg"
    OBJ_NAME = _("image")
    ROI_ITEM_TYPES = (AnnotatedRectangle, AnnotatedCircle, AnnotatedPolygon)

    def get_obj_roi_class(self) -> type[ImageROI]:
        """Get object ROI class"""
        return ImageROI

    def add_tools_to_plot_dialog(self) -> None:
        """Add tools to plot dialog"""
        mgr = self.plot_dialog.get_manager()
        mgr.add_toolbar(self.toolbar, ROI_EDITOR_TOOLBAR_ID)
        mgr.add_tool(ROIRectangleTool, self.obj)
        mgr.add_tool(ROICircleTool, self.obj)
        mgr.add_tool(ROIPolygonTool, self.obj)

    def setup_widget(self) -> None:
        """Setup ROI editor widget"""
        super().setup_widget()
        item: MaskedImageItem = self.plot.get_items(item_type=IImageItemType)[0]
        item.set_mask_visible(False)

    def update_roi_titles(self) -> None:
        """Update ROI annotation titles"""
        super().update_roi_titles()
        for index, roi_item in enumerate(self.roi_items):
            roi_item: AnnotatedRectangle | AnnotatedCircle | AnnotatedPolygon
            roi_item.annotationparam.title = f"ROI{index:02d}"
            roi_item.annotationparam.update_item(roi_item)



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
Object model
============

The :mod:`cdl.core.gui.objectmodel` module defines the object data model used by the
GUI to store signals, images and groups.

The model is based on a hierarchical tree of objects, with two levels:

- The top level is a list of groups (`ObjectGroup` instances)
- The second level is a list of objects (`SignalObj` or `ImageObj` instances)

The model is implemented by the `ObjectModel` class.

Object group
------------

The `ObjectGroup` class represents a group of objects. It is a container for
`SignalObj` and `ImageObj` instances.

.. autoclass:: ObjectGroup

Object model
------------

The `ObjectModel` class is a container for ObjectGroup instances, as well as
a container for `SignalObj` and `ImageObj` instances.

.. autoclass:: ObjectModel
"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...

from __future__ import annotations

import re
from collections.abc import Iterator
from typing import TYPE_CHECKING
from uuid import uuid4

if TYPE_CHECKING:
    from cdl.core.model.image import ImageObj
    from cdl.core.model.signal import SignalObj


# TODO: [P3] Remove this function and use the same mechanism as in reordering methods
#
# This mechanism is as follows (see `reorder_groups` and `reorder_objects` in
# `ObjectModel` class):
# - Replace short IDs by uuids in titles (`__replace_short_ids_by_uuids_in_titles`)
# - Do the operation (object/group removal or any other operation)
# - Reset short IDs (`reset_short_ids`)
# - Replace uuids by short IDs in titles (`__replace_uuids_by_short_ids_in_titles`)
# This would be a cleaner solution.
def fix_titles(
    objlist: list[ObjectGroup | SignalObj | ImageObj],
    obj: ObjectGroup | SignalObj | ImageObj,
    operation: str,
) -> None:
    """Fix all object/group titles before adding or removing an object/group

    Args:
        objlist (list[ObjectGroup | SignalObj | ImageObj]): list of objects/groups
        obj (ObjectGroup | SignalObj | ImageObj): object/group to be added or removed
        operation (str): operation to be performed ("add" or "remove")
    """
    assert len(objlist) > 0
    assert operation in ("add", "remove")
    sign = 1 if operation == "add" else -1
    onb = obj.number
    pfx = obj.PREFIX
    oname = f"{pfx}%03d"
    for obj_i in objlist:
        for match in re.finditer(pfx + "[0-9]{3}", obj_i.title):
            before = match.group()
            i_match = int(before[1:])
            if sign == -1 and i_match == onb:
                after = f"{pfx}xxx"
            elif (sign == -1 and i_match > onb) or (sign == 1 and i_match >= onb):
                after = oname % (i_match + sign)
            else:
                continue
            obj_i.title = obj_i.title.replace(before, after)


class ObjectGroup:
    """Represents a DataLab object group

    Args:
        title: group title
        model: object model
    """

    PREFIX = "g"

    def __init__(self, title: str, model: ObjectModel) -> None:
        self.model = model
        self.uuid: str = str(uuid4())  # Group uuid
        self.__objects: list[str] = []  # list of object uuids
        self.__title: str = title
        self.__gnb = 0

    @property
    def number(self) -> int:
        """Return group number (used for short ID)"""
        return self.__gnb

    @number.setter
    def number(self, gnb: int):
        """Set group number (used for short ID)"""
        self.__gnb = gnb

    @property
    def short_id(self):
        """Short group ID"""
        return f"{self.PREFIX}{self.__gnb:03d}"

    @property
    def title(self) -> str:
        """Return group title"""
        return self.__title

    @title.setter
    def title(self, title: str) -> None:
        """Set group title"""
        self.__title = title

    def __iter__(self) -> Iterator[SignalObj | ImageObj]:
        """Iterate over objects in group"""
        return iter(self.model.get_objects(self.__objects))

    def __len__(self) -> int:
        """Return number of objects in group"""
        return len(self.__objects)

    def __getitem__(
        self, index: int | slice
    ) -> SignalObj | ImageObj | list[SignalObj | ImageObj]:
        """Return object at index"""
        if isinstance(index, slice):
            return [
                self.model[self.__objects[i]]
                for i in range(*index.indices(len(self)))
                if i < len(self)
            ]
        return self.model[self.__objects[index]]

    def __contains__(self, obj: SignalObj | ImageObj) -> bool:
        """Return True if obj is in group"""
        return obj.uuid in self.__objects

    def append(self, obj: SignalObj | ImageObj) -> None:
        """Append object to group"""
        self.__objects.append(obj.uuid)

    def insert(self, index: int, obj: SignalObj | ImageObj) -> None:
        """Insert object at index"""
        fix_titles(self.model.get_all_objects(), obj, "add")
        self.__objects.insert(index, obj.uuid)

    def remove(self, obj: SignalObj | ImageObj) -> None:
        """Remove object from group"""
        fix_titles(self.model.get_all_objects(), obj, "remove")
        self.__objects.remove(obj.uuid)

    def clear(self) -> None:
        """Clear group"""
        self.__objects.clear()

    def get_objects(self) -> list[SignalObj | ImageObj]:
        """Return objects in group"""
        return self.model.get_objects(self.__objects)

    def get_object_ids(self) -> list[str]:
        """Return object ids in group"""
        return self.__objects


class ObjectModel:
    """Represents a DataLab object model (groups of signals/images)"""

    def __init__(self) -> None:
        # dict of objects, key is object uuid:
        self._objects: dict[str, SignalObj | ImageObj] = {}
        # list of groups:
        self._groups: list[ObjectGroup] = []

    def reset_short_ids(self) -> None:
        """Reset short IDs (used for object numbering)

        This method is called when an object was removed from a group."""
        gnb = onb = 1
        for group in self._groups:
            group.number = gnb
            gnb += 1
            for obj in group:
                obj.number = onb
                onb += 1

    def __len__(self) -> int:
        """Return number of objects"""
        return len(self._objects)

    def __getitem__(self, uuid: str) -> SignalObj | ImageObj:
        """Return object with uuid"""
        return self._objects[uuid]

    def __iter__(self) -> Iterator[SignalObj | ImageObj]:
        """Iterate over objects"""
        return iter(self._objects.values())

    def __repr__(self) -> str:
        """Return object representation"""
        return repr(self._objects)

    def __str__(self) -> str:
        """Return object string representation"""
        return str(self._objects)

    def __bool__(self) -> bool:
        """Return True if model is not empty"""
        return bool(self._objects)

    def __contains__(self, obj: SignalObj | ImageObj) -> bool:
        """Return True if obj is in model"""
        return obj.uuid in self._objects

    def clear(self) -> None:
        """Clear model"""
        self._objects.clear()
        self._groups.clear()

    def get_all_objects(self) -> list[SignalObj | ImageObj]:
        """Return all objects, in order of appearance in groups"""
        objects = []
        for group in self._groups:
            objects.extend(group.get_objects())
        return objects

    def get_object_or_group(self, uuid: str) -> SignalObj | ImageObj | ObjectGroup:
        """Return object or group with uuid"""
        if uuid in self._objects:
            return self._objects[uuid]
        for group in self._groups:
            if group.uuid == uuid:
                return group
        raise KeyError(f"Object or group with uuid {uuid} not found")

    def get_group(self, uuid: str) -> ObjectGroup:
        """Return group with uuid"""
        for group in self._groups:
            if group.uuid == uuid:
                return group
        raise KeyError(f"Group with uuid {uuid} not found")

    def get_group_from_number(self, number: int) -> ObjectGroup:
        """Return group from its number.

        Args:
            number: group number (unique in model)

        Returns:
            Group

        Raises:
            IndexError: if group with number not found
        """
        for group in self._groups:
            if group.number == number:
                return group
        raise IndexError(f"Group with number {number} not found")

    def get_group_from_object(self, obj: SignalObj | ImageObj) -> ObjectGroup:
        """Return group containing object

        Args:
            obj: object to find group for

        Returns:
            Group

        Raises:
            KeyError: if object not found in any group
        """
        for group in self._groups:
            if obj in group:
                return group
        raise KeyError(f"Object with uuid '{obj.uuid}' not found in any group")

    def get_groups(self, uuids: list[str] | None = None) -> list[ObjectGroup]:
        """Return groups"""
        if uuids is None:
            return self._groups
        return [group for group in self._groups if group.uuid in uuids]

    def add_group(self, title: str) -> ObjectGroup:
        """Add group to model"""
        group = ObjectGroup(title, self)
        gnb = 1
        if self._groups:
            gnb += self._groups[-1].number
        group.number = gnb
        self._groups.append(group)
        return group

    def get_object_group_id(self, obj: SignalObj | ImageObj) -> str | None:
        """Return group id of object

        Args:
            obj: object to get group id from

        Returns:
            group id or None if object is not in any group
        """
        try:
            return self.get_group_from_object(obj).uuid
        except KeyError:
            return None

    def get_group_object_ids(self, group_id: str) -> list[str]:
        """Return object ids in group"""
        for group in self._groups:
            if group.uuid == group_id:
                return group.get_object_ids()
        raise KeyError(f"Group with uuid '{group_id}' not found")

    def remove_group(self, group: ObjectGroup) -> None:
        """Remove group from model"""
        fix_titles(self.get_groups(), group, "remove")
        self._groups.remove(group)
        for obj in group:
            remove_obj = True
            for other_group in self._groups:
                if obj in other_group:
                    remove_obj = False
                    break
            if remove_obj:
                del self._objects[obj.uuid]
        self.reset_short_ids()

    def add_object(self, obj: SignalObj | ImageObj, group_id: str) -> None:
        """Add object to model"""
        self.__replace_short_ids_by_uuids_in_titles([obj])
        self._objects[obj.uuid] = obj
        onb = 0
        for group in self._groups:
            onb += len(group)
            if group.uuid == group_id:
                obj.number = onb + 1
                group.append(obj)
                break
        else:
            raise KeyError(f"Group with uuid '{group_id}' not found")
        self.reset_short_ids()
        self.__replace_uuids_by_short_ids_in_titles()

    def remove_object(self, obj: SignalObj | ImageObj) -> None:
        """Remove object from model"""
        for group in self._groups:
            if obj in group:
                group.remove(obj)
        del self._objects[obj.uuid]
        self.reset_short_ids()

    def get_object_from_number(self, number: int) -> SignalObj | ImageObj:
        """Return object from its number.

        Args:
            number: object number (unique in model)

        Returns:
            Object

        Raises:
            IndexError: if object with number not found
        """
        for obj in self._objects.values():
            if obj.number == number:
                return obj
        raise IndexError(f"Object with number {number} not found")

    def get_objects(self, uuids: list[str]) -> list[SignalObj | ImageObj]:
        """Return objects with uuids"""
        return [self._objects[uuid] for uuid in uuids]

    def get_object_ids(self) -> list[str]:
        """Return object ids, in order of appearance in groups"""
        return [obj.uuid for obj in self.get_all_objects()]

    def get_group_titles_with_object_infos(
        self,
    ) -> tuple[list[str], list[list[str]], list[list[str]]]:
        """Return groups titles and lists of inner objects uuids and titles.

        Returns:
            Tuple: groups titles, lists of inner objects uuids and titles
        """
        grp_titles = []
        obj_uuids = []
        obj_titles = []
        for group in self._groups:
            grp_titles.append(group.title)
            obj_uuids.append(group.get_object_ids())
            obj_titles.append([obj.title for obj in group])
        return grp_titles, obj_uuids, obj_titles

    def get_object_titles(self) -> list[str]:
        """Return object titles, in order of appearance in groups"""
        return [obj.title for obj in self.get_all_objects()]

    def get_object_from_title(self, title: str) -> SignalObj | ImageObj:
        """Return object with title.

        Args:
            title: object title

        Returns:
            object with title

        Raises:
            KeyError: if object with title not found
        """
        for obj in self._objects.values():
            if obj.title == title:
                return obj
        raise KeyError(f"Object with title '{title}' not found")

    def __get_group_object_mapping_to_shortid(self) -> dict[str, str]:
        """Return dictionary mapping group/object uuids to their short ID"""
        mapping = {}
        for group in self._groups:
            mapping[group.uuid] = group.short_id
            for obj in group:
                mapping[obj.uuid] = obj.short_id
        return mapping

    def __replace_short_ids_by_uuids_in_titles(
        self, other_objects: tuple[SignalObj | ImageObj] | None = None
    ) -> None:
        """Replace short IDs by uuids in titles

        Args:
            other_objects: tuple of other objects to consider for short ID replacement

        .. note::

            This method is called before reorganizing groups or objects. It replaces the
            short IDs in titles by the uuids. This is needed because the short IDs are
            used to reflect in the title the operation performed on the object/group,
            e.g. "fft(s001)" or "g001 + g002". But when reorganizing groups or objects,
            the short IDs may change, so we need to replace them by the uuids, which are
            stable. Once the reorganization is done, we will replace the uuids by the
            new short IDs thanks to the `__replace_uuids_by_short_ids_in_titles` method.
        """
        mapping = self.__get_group_object_mapping_to_shortid()
        objs = self._objects.values()
        if other_objects is not None:
            objs = list(objs) + list(other_objects)
        for obj in objs:
            for obj_uuid, short_id in mapping.items():
                obj.title = obj.title.replace(short_id, obj_uuid)
        for group in self._groups:
            for grp_uuid, short_id in mapping.items():
                group.title = group.title.replace(short_id, grp_uuid)

    def __replace_uuids_by_short_ids_in_titles(self) -> None:
        """Replace uuids by short IDs in titles

        .. note::

            This method is called after reorganizing groups or objects. It replaces
            the uuids in titles by the short IDs.
        """
        mapping = self.__get_group_object_mapping_to_shortid()
        for obj in self._objects.values():
            for obj_uuid, short_id in mapping.items():
                obj.title = obj.title.replace(obj_uuid, short_id)
        for group in self._groups:
            for grp_uuid, short_id in mapping.items():
                group.title = group.title.replace(grp_uuid, short_id)
        # Replace remaining UUIDs with f"{obj.PREFIX}xxx"
        # (this may happen if groups or objects were removed in the meantime):
        pattern = re.compile(r"\b[0-9a-f]{8}-([0-9a-f]{4}-){3}[0-9a-f]{12}\b")
        for obj in self._objects.values():
            obj.title = pattern.sub(f"{obj.PREFIX}xxx", obj.title)
        for group in self._groups:
            for obj in group:
                obj.title = pattern.sub(f"{obj.PREFIX}xxx", obj.title)

    def reorder_groups(self, group_ids: list[str]) -> None:
        """Reorder groups.

        Args:
            group_ids: list of group uuids
        """
        # Replace short IDs by uuids in titles:
        self.__replace_short_ids_by_uuids_in_titles()
        # Reordering groups:
        self._groups = [self.get_group(group_id) for group_id in group_ids]
        # Reset short IDs:
        self.reset_short_ids()
        # Replace uuids by short IDs in titles:
        self.__replace_uuids_by_short_ids_in_titles()

    def reorder_objects(self, obj_ids: dict[str, list[str]]) -> None:
        """Reorder objects in groups.

        Args:
            obj_ids: dict of group uuids and list of object uuids
        """
        # Replace short IDs by uuids in titles:
        self.__replace_short_ids_by_uuids_in_titles()
        # Reordering objects in groups:
        for group_id, obj_uuids in obj_ids.items():
            group = self.get_group(group_id)
            group.clear()
            for obj_uuid in obj_uuids:
                group.append(self._objects[obj_uuid])
        # Reset short IDs:
        self.reset_short_ids()
        # Replace uuids by short IDs in titles:
        self.__replace_uuids_by_short_ids_in_titles()



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
Unit tests for image data types
-------------------------------

"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...

from __future__ import annotations

import numpy as np

import cdl.obj
from cdl.algorithms.datatypes import clip_astype, is_integer_dtype
from cdl.env import execenv


def get_integer_datatypes() -> list[cdl.obj.ImageDatatypes]:
    """Return all integer data types."""
    return [dtype for dtype in cdl.obj.ImageDatatypes if "int" in dtype.name.lower()]


def test_clip_astype():
    """Test `clip_astype` algorithm"""

    # Test that function do nothing for certain data types
    for dtype1 in cdl.obj.ImageDatatypes:
        for dtype2 in cdl.obj.ImageDatatypes:
            if not is_integer_dtype(dtype2.value):
                continue
            if is_integer_dtype(dtype1.value):
                info1 = np.iinfo(dtype1.value)
            else:
                info1 = np.finfo(dtype1.value)
            info2 = np.iinfo(dtype2.value)
            if info2.min < info1.min or info2.max > info1.max:
                continue
            data = np.array([0, 1, 2, 3, 4, 5], dtype=dtype1.value)
            txt = f"No change: {dtype1.value} -> {dtype2.value}"
            execenv.print(txt, end="... ")
            assert np.array_equal(clip_astype(data, dtype2.value), data), txt
            execenv.print("OK")

    # Test that function handles overflow for integer data types
    for dtype in get_integer_datatypes():
        maxval = np.iinfo(dtype.value).max
        data1 = np.array([maxval], dtype=dtype.value)
        data2 = clip_astype(data1.astype(float) + 1, dtype.value)
        txt = f"Overflow: {dtype.value}"
        execenv.print(txt, end="... ")
        assert data2[0] == maxval, txt
        execenv.print("OK")

    # Test that function handles underflow for integer data types
    for dtype in get_integer_datatypes():
        minval = np.iinfo(dtype.value).min
        data1 = np.array([minval], dtype=dtype.value)
        data2 = clip_astype(data1.astype(float) - 1, dtype.value)
        txt = f"Underflow: {dtype.value}"
        execenv.print(txt, end="... ")
        assert data2[0] == minval, txt
        execenv.print("OK")


if __name__ == "__main__":
    test_clip_astype()



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
Statistics test

Testing the following:
  - Create a signal
  - Compute statistics on signal and show results
  - Create an image
  - Compute statistics on image and show results
"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...
# guitest: show

from cdl.tests import cdltest_app_context, take_plotwidget_screenshot
from cdl.tests.features.common.stat_unit_test import (
    create_reference_image,
    create_reference_signal,
)


def test_stat_app():
    """Run statistics application test scenario"""
    with cdltest_app_context() as win:
        # === Signal statistics test ===
        panel = win.signalpanel
        panel.add_object(create_reference_signal())
        take_plotwidget_screenshot(panel, "stat_test")
        panel.processor.compute_stats()
        # === Image statistics test ===
        panel = win.imagepanel
        panel.add_object(create_reference_image())
        take_plotwidget_screenshot(panel, "stat_test")
        panel.processor.compute_stats()


if __name__ == "__main__":
    test_stat_app()



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
Update processor methods CSV files
----------------------------------

The purpose of this script is to generate CSV files containing the list of compute
methods attached to the processor classes:

- :class:`cdl.core.gui.processor.signal.SignalProcessor`
- :class:`cdl.core.gui.processor.image.ImageProcessor`

Those methods are *almost* associated elementwise with the compute functions defined in
the :mod:`cdl.computation` package.
"""

from __future__ import annotations

import csv
import inspect
import os.path as osp

from cdl.core.gui.processor.image import ImageProcessor
from cdl.core.gui.processor.signal import SignalProcessor
from cdl.utils.strings import shorten_docstring


def get_compute_methods(klass: type) -> list[str]:
    """Retrieve list of `compute_` methods from a processor class

    Args:
        klass: Processor class

    Returns:
        List of tuples containing the function name, and docstring
    """
    compute_methods = []
    for name, obj in inspect.getmembers(klass, inspect.isfunction):
        if name.startswith("compute_") and name not in (
            "compute_10",
            "compute_11",
            "compute_1n",
            "compute_n1",
            "compute_n1n",
        ):
            compute_methods.append((name, obj.__doc__))
    return compute_methods


def generate_csv_files() -> None:
    """Generate CSV files containing the validation status of compute functions"""
    lengths = []
    for category, klass in (("signal", SignalProcessor), ("image", ImageProcessor)):
        rows = []
        methods = get_compute_methods(klass)
        lengths.append(len(methods))
        for name, docstring in methods:
            pyfunc_link = f":py:func:`~{klass.__module__}.{klass.__name__}.{name}`"
            rows.append([pyfunc_link, shorten_docstring(docstring)])
        fname = osp.join(osp.dirname(__file__), f"processor_methods_{category}.csv")
        with open(fname, "w", newline="", encoding="utf-8") as csvfile:
            writer = csv.writer(csvfile)
            writer.writerows(rows)
    lengths.append(sum(lengths))

    fname = osp.join(osp.dirname(__file__), "processor_methods_nb.csv")
    with open(fname, "w", newline="", encoding="utf-8") as csvfile:
        writer = csv.writer(csvfile)
        writer.writerows([[str(length) for length in lengths]])

    # Print statistics:
    print("Processor methods:")
    print(f"  Signal: {lengths[0]}")
    print(f"  Image:  {lengths[1]}")
    print(f"  Total:  {lengths[2]}")
    print()


if __name__ == "__main__":
    generate_csv_files()



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
Annotations unit test:

  - Create an image with annotations
  - Open dialog (equivalent to click on button "Annotations")
  - Accept dialog without modifying shapes
  - Check if image annotations are still the same
"""

# guitest: show

from plotpy.builder import make
from plotpy.items import AnnotatedShape, PolygonShape
from plotpy.plot import BasePlot
from qtpy import QtWidgets as QW

from cdl.core.model.base import ANN_KEY
from cdl.env import execenv
from cdl.tests import cdltest_app_context
from cdl.tests import data as test_data


def set_annotation_color(annotation: AnnotatedShape, color: str) -> None:
    """Set annotation color"""
    shape: PolygonShape = annotation.shape
    param = shape.shapeparam
    param.line.color = param.fill.color = color
    param.fill.alpha = 0.3
    param.fill.style = "SolidPattern"
    param.update_item(shape)
    plot: BasePlot = annotation.plot()
    if plot is not None:
        plot.replot()


def test_annotations_unit():
    """Run image tools test scenario"""
    with cdltest_app_context() as win:
        panel = win.imagepanel

        # Create image with annotations
        ima1 = test_data.create_multigauss_image()
        ima1.title = "Annotations from items"
        rect = make.annotated_rectangle(100, 100, 200, 200, title="Test")
        set_annotation_color(rect, "#2222ff")
        circ = make.annotated_circle(300, 300, 400, 400, title="Test")
        set_annotation_color(circ, "#22ff22")
        elli = make.annotated_ellipse(
            500, 500, 800, 500, 650, 400, 650, 600, title="Test"
        )
        segm = make.annotated_segment(700, 700, 800, 800, title="Test")
        label = make.label("Test", (1000, 1000), (0, 0), "BR")
        ima1.add_annotations_from_items([rect, circ, elli, segm, label])
        panel.add_object(ima1)

        # Create another image with annotations
        ima2 = test_data.create_annotated_image(title="Annotations from JSON")
        panel.add_object(ima2)

        execenv.print("Check [geometric shapes] <--> [plot items] conversion:")
        execenv.print(f"(comparing {ANN_KEY} metadata)")
        for ima in (ima1, ima2):
            execenv.print(f"  Checking image '{ima.title}': ", end="")
            panel.objview.select_objects([ima])
            # Open separate view
            panel.open_separate_view().done(QW.QDialog.DialogCode.Accepted)
            orig_metadata = ima.metadata.copy()
            panel.open_separate_view().done(QW.QDialog.DialogCode.Accepted)
            # Check if metadata are still the same
            # open("before.json", mode="wb").write(orig_metadata[ANN_KEY].encode())
            # open("after.json", mode="wb").write(ima.metadata[ANN_KEY].encode())
            assert orig_metadata[ANN_KEY] == ima.metadata[ANN_KEY]
            execenv.print("OK")


if __name__ == "__main__":
    test_annotations_unit()



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
Process isolation unit test
---------------------------

Just test if it's possible to run the process isolation test from another module. This
may be an issue with the Pool object being global.
"""

# guitest: show

from cdl.tests.backbone import procisolation1_unit


def test_procisolation():
    """Test process isolation"""
    procisolation1_unit.test_multiprocessing1()


if __name__ == "__main__":
    test_procisolation()



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
Contour finding test
"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...
# pylint: disable=duplicate-code
# guitest: show

import sys
import time

from guidata.qthelpers import qt_app_context
from plotpy.builder import make

from cdl.algorithms import coordinates
from cdl.algorithms.image import get_2d_peaks_coords, get_contour_shapes
from cdl.env import execenv
from cdl.tests.data import get_peak2d_data
from cdl.utils.vistools import view_image_items


def find_contours(data):
    """Find contours"""
    items = [make.image(data, interpolation="linear", colormap="hsv")]
    t0 = time.time()
    peak_coords = get_2d_peaks_coords(data)
    dt = time.time() - t0
    for x, y in peak_coords:
        items.append(make.marker((x, y)))
    execenv.print(f"Calculation time: {int(dt * 1e3):d} ms\n", file=sys.stderr)
    execenv.print(f"Peak coordinates: {peak_coords}")
    for shape in ("circle", "ellipse", "polygon"):
        coords = get_contour_shapes(data, shape=shape)
        execenv.print(f"Coordinates ({shape}s): {coords}")
        for shapeargs in coords:
            if shape == "circle":
                xc, yc, r = shapeargs
                x0, y0, x1, y1 = coordinates.circle_to_diameter(xc, yc, r)
                item = make.circle(x0, y0, x1, y1)
            elif shape == "ellipse":
                xc, yc, a, b, theta = shapeargs
                coords = coordinates.ellipse_to_diameters(xc, yc, a, b, theta)
                x0, y0, x1, y1, x2, y2, x3, y3 = coords
                item = make.ellipse(x0, y0, x1, y1, x2, y2, x3, y3)
            else:
                # `shapeargs` is a flattened array of x, y coordinates
                x, y = shapeargs[::2], shapeargs[1::2]
                item = make.polygon(x, y, closed=False)
            items.append(item)
    view_image_items(items)


def test_contour():
    """2D peak detection test"""
    with qt_app_context():
        find_contours(get_peak2d_data())


if __name__ == "__main__":
    test_contour()



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
Signal spectrum unit test.
"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...
# pylint: disable=duplicate-code
# guitest: show

from __future__ import annotations

from guidata.qthelpers import qt_app_context

import cdl.algorithms.signal as alg
from cdl.tests.data import get_test_signal
from cdl.utils.vistools import view_curves


def test_signal_magnitude_spectrum_interactive() -> None:
    """Interactive test of the magnitude spectrum of a signal."""
    with qt_app_context():
        obj = get_test_signal("dynamic_parameters.txt")
        x, y = obj.xydata
        xms, yms = alg.magnitude_spectrum(x, y, log_scale=True)
        view_curves(
            [(xms, yms)],
            title="Magnitude spectrum",
            xlabel="Frequency",
            ylabel="Magnitude",
        )


def test_signal_phase_spectrum_interactive() -> None:
    """Interactive test of the phase spectrum of a signal."""
    with qt_app_context():
        obj = get_test_signal("dynamic_parameters.txt")
        x, y = obj.xydata
        xps, yps = alg.phase_spectrum(x, y)
        view_curves(
            [(xps, yps)],
            title="Phase spectrum",
            xlabel="Frequency",
            ylabel="Phase",
        )


def test_signal_psd_interactive() -> None:
    """Interactive test of the power spectral density of a signal."""
    with qt_app_context():
        obj = get_test_signal("dynamic_parameters.txt")
        x, y = obj.xydata
        xpsd, ypsd = alg.psd(x, y, log_scale=True)
        view_curves(
            [(xpsd, ypsd)],
            title="Power spectral density",
            xlabel="Frequency",
            ylabel="Power",
        )


if __name__ == "__main__":
    test_signal_magnitude_spectrum_interactive()
    test_signal_phase_spectrum_interactive()
    test_signal_psd_interactive()



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
Process isolation unit test
---------------------------
"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...
# pylint: disable=duplicate-code
# guitest: skip

from __future__ import annotations

import time
from collections.abc import Callable
from multiprocessing.pool import AsyncResult, Pool
from typing import Any

import numpy as np
import scipy.signal as sps
from guidata.qthelpers import qt_app_context
from plotpy.plot import PlotWindow
from qtpy import QtWidgets as QW

import cdl.obj
from cdl.env import execenv
from cdl.tests.data import create_2d_random, create_noisygauss_image
from cdl.utils.qthelpers import create_progress_bar

POOL: Pool | None = None


class Worker:
    """Multiprocessing worker, to run long-running tasks in a separate process"""

    def __init__(self) -> None:
        self.asyncresult: AsyncResult = None
        self.result: Any = None

    @staticmethod
    def create_pool() -> None:
        """Create multiprocessing pool"""
        global POOL  # pylint: disable=global-statement
        # Create a pool with one process
        POOL = Pool(processes=1)  # pylint: disable=not-callable,consider-using-with

    @staticmethod
    def terminate_pool() -> None:
        """Terminate multiprocessing pool"""
        global POOL  # pylint: disable=global-statement
        if POOL is not None:
            POOL.terminate()
            POOL.join()
            POOL = None

    def run(self, func: Callable, args: tuple[Any]) -> None:
        """Run computation"""
        global POOL  # pylint: disable=global-statement,global-variable-not-assigned
        assert POOL is not None
        self.asyncresult = POOL.apply_async(func, args)

    def terminate(self) -> None:
        """Terminate worker"""
        # Terminate the process and stop the timer
        self.terminate_pool()
        execenv.print("Computation cancelled!")
        # Recreate the pool for the next computation
        self.create_pool()

    def is_computation_finished(self) -> bool:
        """Return True if computation is finished"""
        return self.asyncresult.ready()

    def get_result(self) -> Any:
        """Return computation result"""
        self.result = self.asyncresult.get()
        self.asyncresult = None
        return self.result


def testfunc(data: np.ndarray, size: int, error: bool) -> np.ndarray:
    """Test function"""
    if error:
        raise ValueError("Test error")
    return sps.medfilt(data, size) + create_2d_random(data.shape[0], data.dtype)


def test_multiprocessing1(iterations: int = 4) -> None:
    """Multiprocessing test"""
    Worker.create_pool()
    with qt_app_context(exec_loop=True):
        win = PlotWindow(title="Multiprocessing test", icon="datalab.svg", toolbar=True)
        win.resize(800, 600)
        win.show()
        param = cdl.obj.new_image_param(
            height=1000, width=1000, dtype=cdl.obj.ImageDatatypes.UINT16
        )
        image = create_noisygauss_image(param, add_annotations=True)
        win.get_plot().add_item(image.make_item())
        worker = Worker()
        with create_progress_bar(win, "Computing", max_=iterations) as progress:
            for index in range(iterations):
                progress.setValue(index)
                progress.setLabelText(f"Computing {index}")
                test_error = index == 2
                worker.run(testfunc, (image.data, 3, test_error))
                while not worker.is_computation_finished():
                    QW.QApplication.processEvents()
                    time.sleep(0.1)
                    if progress.wasCanceled():
                        worker.terminate()
                        break
                if worker.is_computation_finished():
                    try:
                        image.data = worker.get_result()
                    except Exception as exc:  # pylint: disable=broad-except
                        execenv.print(f"Intercepted exception: {exc}")
                    win.get_plot().add_item(image.make_item())
                else:
                    break
        worker.terminate_pool()


if __name__ == "__main__":
    test_multiprocessing1()



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
Miscellaneous application test
------------------------------

Whenever we have a test that does not fit in any of the other test files,
we put it here...
"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...
# guitest: show

from __future__ import annotations

from typing import TYPE_CHECKING, Any

import cdl.obj
import cdl.param
from cdl.env import execenv
from cdl.tests import cdltest_app_context
from cdl.tests.data import (
    create_2dstep_image,
    create_paracetamol_signal,
    get_test_fnames,
)

if TYPE_CHECKING:
    from cdl.core.gui.main import CDLMainWindow


def __print_test_result(title: str, result: Any = None) -> None:
    """Print a test result"""
    execenv.print(f"Testing {title}{'' if result is None else ':'}")
    if result is not None:
        execenv.print(str(result))
        execenv.print("")


def __misc_unit_function(win: CDLMainWindow) -> None:
    """Run miscellaneous unit tests

    Args:
        win: CDLMainWindow instance
    """
    panel = win.signalpanel
    objview = panel.objview

    sig = create_paracetamol_signal()
    panel.add_object(sig)
    panel.processor.compute_derivative()
    panel.processor.compute_moving_average(cdl.param.MovingAverageParam.create(n=5))

    __print_test_result("`SimpleObjectTree.__str__` method", objview)

    # Updated metadata view settings:
    __print_test_result("Updated metadata view settings")
    panel.update_metadata_view_settings()

    # Double click on the first signal item in object view:
    __print_test_result("Double click on the first signal item in object view")
    tree_item = objview.currentItem()
    objview.itemDoubleClicked.emit(tree_item, 0)

    # Open context menu on current item:
    __print_test_result("Open context menu on current item")
    tree_item_pos = objview.mapToGlobal(objview.visualItemRect(tree_item).center())
    objview.SIG_CONTEXT_MENU.emit(tree_item_pos)

    # Plot item parameters changed:
    __print_test_result("Plot item parameters changed")
    objview.select_objects([sig.uuid])
    item = panel.plothandler[sig.uuid]
    panel.plot_item_parameters_changed(item)

    # Duplicate group:
    __print_test_result("Duplicate group")
    objview.select_groups([1])
    panel.duplicate_object()

    # Delete group:
    __print_test_result("Delete group")
    objview.select_groups([2])
    panel.remove_object()

    # Exec import wizard:
    __print_test_result("Exec signal import wizard")
    win.signalpanel.exec_import_wizard()
    __print_test_result("Exec image import wizard")
    win.imagepanel.exec_import_wizard()

    # Properties changed:
    __print_test_result("Properties changed")
    objview.select_objects([sig.uuid])
    panel.properties_changed()

    # Get object titles:
    __print_test_result("Get object titles")
    execenv.print(win.get_object_titles())

    # Get object titles with info:
    __print_test_result("Get group titles with object infos")
    execenv.print(win.get_group_titles_with_object_infos())

    # Pop up tab menu:
    __print_test_result("Pop up tab menu")
    win.tabmenu.popup(win.mapToGlobal(win.tabmenu.pos()))

    # Repopulate panel trees:
    __print_test_result("Repopulate panel trees")
    win.repopulate_panel_trees()

    # Browse HDF5 files:
    __print_test_result("Browse HDF5 files")
    win.browse_h5_files([], False)

    # Open object
    __print_test_result("Open object")
    fname = get_test_fnames("*.csv")[0]
    win.load_from_files([fname])

    # Open objects from signal panel
    __print_test_result("Open objects from signal panel")
    win.signalpanel.load_from_files(get_test_fnames("curve_formats/*.*"))

    # Get version
    __print_test_result("Get version")
    execenv.print(win.get_version())

    # Add signal
    __print_test_result("Add signal")
    win.add_signal(
        sig.title, sig.x, sig.y, sig.xunit, sig.yunit, sig.xlabel, sig.ylabel
    )

    # Add image
    __print_test_result("Add image")
    ima = create_2dstep_image()
    win.add_image(
        ima.title,
        ima.data,
        ima.xunit,
        ima.yunit,
        ima.zunit,
        ima.xlabel,
        ima.ylabel,
        ima.zlabel,
    )

    # Close application
    __print_test_result("Close application")
    win.close_application()


def test_misc_app() -> None:
    """Run misc. application test scenario"""
    with cdltest_app_context(console=False) as win:
        execenv.print("Miscellaneous application test")
        execenv.print("==============================")
        execenv.print("")

        # We don't need to refresh the GUI during those tests (that's faster!)
        with win.context_no_refresh():
            __misc_unit_function(win)


if __name__ == "__main__":
    test_misc_app()



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
DataLab pytest configuration
----------------------------

This file contains the configuration for running pytest in DataLab. It is
executed before running any tests.
"""

import os

import guidata
import h5py
import numpy
import plotpy
import pytest
import qtpy
import qwt
import scipy
import skimage

import cdl
from cdl.env import execenv
from cdl.plugins import get_available_plugins

# Turn on unattended mode for executing tests without user interaction
execenv.unattended = True
execenv.verbose = "quiet"

INITIAL_CWD = os.getcwd()


def pytest_report_header(config):  # pylint: disable=unused-argument
    """Add additional information to the pytest report header."""
    nfstr = ", ".join(
        f"{plugin.info.name} {plugin.info.version}"
        for plugin in get_available_plugins()
    )
    qtbindings_version = qtpy.PYSIDE_VERSION
    if qtbindings_version is None:
        qtbindings_version = qtpy.PYQT_VERSION
    infolist = [
        f"DataLab {cdl.__version__} [Plugins: {nfstr if nfstr else 'None'}]",
        f"guidata {guidata.__version__}, PlotPy {plotpy.__version__}",
        f"PythonQwt {qwt.__version__}, "
        f"{qtpy.API_NAME} {qtbindings_version} [Qt version: {qtpy.QT_VERSION}]",
        f"NumPy {numpy.__version__}, SciPy {scipy.__version__}, "
        f"h5py {h5py.__version__}, scikit-image {skimage.__version__}",
    ]
    try:
        import cv2  # pylint: disable=import-outside-toplevel

        infolist[-1] += f", OpenCV {cv2.__version__}"
    except ImportError:
        pass
    for vname in ("CDL_DATA", "PYTHONPATH", "DEBUG"):
        value = os.environ.get(vname, "")
        if value:
            infolist.append(f"{vname}: {value}")
    return infolist


def pytest_configure(config):
    """Add custom markers to pytest."""
    config.addinivalue_line(
        "markers",
        "validation: mark a test as a validation test (ground truth or analytical)",
    )


@pytest.fixture(autouse=True)
def reset_cwd(request):  # pylint: disable=unused-argument
    """Reset the current working directory to the initial one after each test."""
    yield
    os.chdir(INITIAL_CWD)



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
DataLab
=======

DataLab is a generic signal and image processing software based on Python
scientific libraries (such as NumPy, SciPy or scikit-image) and Qt graphical
user interfaces (thanks to `PlotPyStack`_ libraries).

.. _PlotPyStack: https://github.com/PlotPyStack
"""

import os

__version__ = "0.18.1"
__docurl__ = __homeurl__ = "https://datalab-platform.com/"
__supporturl__ = "https://github.com/DataLab-Platform/DataLab/issues/new/choose"

os.environ["CDL_VERSION"] = __version__

# Dear (Debian, RPM, ...) package makers, please feel free to customize the
# following path to module's data (images) and translations:
DATAPATH = LOCALEPATH = ""



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
Module providing a log viewer widget, a log viewer window and DataLab's log viewer
"""

from __future__ import annotations

import os.path as osp

from guidata.configtools import get_icon
from guidata.qthelpers import exec_dialog
from qtpy import QtWidgets as QW

from cdl.config import APP_NAME, Conf, _, get_old_log_fname
from cdl.env import execenv
from cdl.widgets.fileviewer import FileViewerWidget, get_title_contents


class LogViewerWindow(QW.QDialog):
    """Log viewer window"""

    def __init__(self, fnames: list[str], parent: QW.QWidget | None = None) -> None:
        super().__init__(parent)
        self.setObjectName("logviewer")
        self.setWindowTitle(APP_NAME + " - " + _("Log files"))
        self.setWindowIcon(get_icon("DataLab.svg"))
        self.tabs = QW.QTabWidget(self)
        for fname in fnames:
            if osp.isfile(fname):
                title, contents = get_title_contents(fname)
                if not contents.strip():
                    continue
                viewer = FileViewerWidget(language="Python")
                viewer.set_data(title, contents)
                self.tabs.addTab(viewer, get_icon("logs.svg"), osp.basename(fname))
        layout = QW.QVBoxLayout()
        layout.addWidget(self.tabs)
        self.setLayout(layout)
        self.resize(900, 400)

    @property
    def is_empty(self) -> bool:
        """Return True if there is no log available"""
        return self.tabs.count() == 0


def get_log_filenames() -> list[str]:
    """Return log filenames"""
    return [
        Conf.main.traceback_log_path.get(),
        Conf.main.faulthandler_log_path.get(),
        get_old_log_fname(Conf.main.traceback_log_path.get()),
        get_old_log_fname(Conf.main.faulthandler_log_path.get()),
    ]


def get_log_prompt_message() -> str | None:
    """Return prompt message for log files, i.e. a message informing the user
    whether log files were generated during last session or current session."""
    avail = [osp.isfile(fname) for fname in get_log_filenames()]
    if avail[0] or avail[1]:
        return _("Log files were generated during current session.")
    if avail[2] or avail[3]:
        return _("Log files were generated during last session.")
    return None


def exec_cdl_logviewer_dialog(parent: QW.QWidget | None = None) -> None:
    """View DataLab logs"""
    fnames = [osp.normpath(fname) for fname in get_log_filenames() if osp.isfile(fname)]
    dlg = LogViewerWindow(fnames, parent=parent)
    if dlg.is_empty:
        if not execenv.unattended:
            QW.QMessageBox.information(
                dlg, APP_NAME, _("Log files are currently empty.")
            )
        dlg.close()
    else:
        exec_dialog(dlg)



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
DataLab I/O signal functions
"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...

from __future__ import annotations

import re
from typing import TextIO

import numpy as np
import pandas as pd

from cdl.utils.io import count_lines
from cdl.utils.qthelpers import CallbackWorker


def get_labels_units_from_dataframe(
    df: pd.DataFrame,
) -> tuple[str, list[str], str, list[str]]:
    """Get labels and units from a DataFrame.

    Args:
        df: DataFrame

    Returns:
        Tuple (xlabel, ylabels, xunit, yunits)
    """
    # Reading X,Y labels
    xlabel = str(df.columns[0])
    ylabels = [str(col) for col in df.columns[1:]]

    # Retrieving units from labels
    xunit = ""
    yunits = [""] * len(ylabels)
    pattern = r"([\S ]*) \(([\S]*)\)"
    match = re.match(pattern, xlabel)
    if match is not None:
        xlabel, xunit = match.groups()
    for i, ylabel in enumerate(ylabels):
        match = re.match(pattern, ylabel)
        if match is not None:
            ylabels[i], yunits[i] = match.groups()

    return xlabel, ylabels, xunit, yunits


def read_csv_by_chunks(
    fname_or_fileobj: str | TextIO,
    nlines: int | None = None,
    worker: CallbackWorker | None = None,
    delimiter: str | None = None,
    header: int | None = "infer",
    skiprows: int | None = None,
    nrows: int | None = None,
    comment: str | None = None,
    chunksize: int = 1000,
) -> pd.DataFrame:
    """Read CSV data with primitive options, using pandas read_csv function defaults,
    and reading data in chunks, using the iterator interface.

    Args:
        fname_or_fileobj: CSV file name or text stream object
        nlines: Number of lines contained in file (this argument is mandatory if
         `fname_or_fileobj` is a text stream object: counting line numbers from a
         text stream is not efficient, especially if one already has access to the
         initial text content from which the text stream was made)
        worker: Callback worker object
        delimiter: Delimiter
        header: Header line
        skiprows: Skip rows
        nrows: Number of rows to read
        comment: Comment character
        chunksize: Chunk size

    Returns:
        DataFrame
    """
    if isinstance(fname_or_fileobj, str):
        nlines = count_lines(fname_or_fileobj)
    elif nlines is None:
        raise ValueError("Argument `nlines` must be passed for text streams")
    # Read data in chunks, and concatenate them at the end, thus allowing to call the
    # progress callback function at each chunk read and to return an intermediate result
    # if the operation is canceled.
    chunks = []
    for chunk in pd.read_csv(
        fname_or_fileobj,
        delimiter=delimiter,
        header=header,
        skiprows=skiprows,
        nrows=nrows,
        comment=comment,
        chunksize=chunksize,
    ):
        chunks.append(chunk)
        # Compute the progression based on the number of lines read so far
        if worker is not None:
            worker.set_progress(sum(len(chunk) for chunk in chunks) / nlines)
            if worker.was_canceled():
                break
    return pd.concat(chunks)


def read_csv(
    filename: str,
    worker: CallbackWorker | None = None,
) -> tuple[
    np.ndarray, str | None, str | None, list[str] | None, list[str] | None, str | None
]:
    """Read CSV data, and return tuple (xydata, xlabel, xunit, ylabels, yunits, header).

    Args:
        filename: CSV file name
        worker: Callback worker object

    Returns:
        Tuple (xydata, xlabel, xunit, ylabels, yunits, header)
    """
    xydata, xlabel, xunit, ylabels, yunits, header = None, None, None, None, None, None

    # The first attempt is to read the CSV file assuming it has no header because it
    # won't raise an error if the first line is data. If it fails, we try to read it
    # with a header, and if it fails again, we try to skip some lines before reading
    # the data.

    skiprows = None

    # First attempt: no header (try to read with different delimiters)
    for delimiter in (",", ";", "\t", " "):
        try:
            df = pd.read_csv(
                filename,
                dtype=float,
                delimiter=delimiter,
                header=None,
                comment="#",
                nrows=1000,  # Read only the first 1000 lines
            )
            read_without_header = True
            break
        except (pd.errors.ParserError, ValueError):
            df = None

    # Second attempt: with header
    if df is None:
        for delimiter in (",", ";", "\t", " "):
            # Headers are generally in the first 10 lines, so we try to skip the
            # minimum number of lines before reading the data:
            for skiprows in range(20):
                try:
                    df = pd.read_csv(
                        filename,
                        dtype=float,
                        delimiter=delimiter,
                        skiprows=skiprows,
                        comment="#",
                        nrows=1000,  # Read only the first 1000 lines
                    )
                    break
                except (pd.errors.ParserError, ValueError):
                    df = None
            if df is not None:
                break

        if df is None:
            raise ValueError("Unable to read CSV file (format not supported)")

        # At this stage, we have a DataFrame with column names, but we don't know
        # if the first line is a header or data. We try to read the first line as
        # a header, and if it fails, we read it as data.
        try:
            df.columns.astype(float)
            # This means that the first line is data, so we have to read it again, but
            # without the header:
            read_without_header = True
        except ValueError:
            read_without_header = False
            # This means that the first line is a header, so we already have the data
            # without missing values.
            # However, it also means that there could be text information preceding
            # the header. Let's try to read it and put it in `header` variable.

            # 1. We read only the first 1000 lines to avoid reading the whole file
            # 2. We keep only the lines beginning with a comment character
            # 3. We join the lines to create a single string
            header = ""
            with open(filename, "r", encoding="utf-8") as file:
                for _ in range(1000):
                    line = file.readline()
                    if line.startswith("#"):
                        header += line
                    else:
                        break
            # Remove the last line if it contains the column names:
            if header and df.columns[0] in header.splitlines()[-1]:
                header = "\n".join(header.splitlines()[:-1])

    # Now we read the whole file with the correct options
    df = read_csv_by_chunks(
        filename,
        worker=worker,
        delimiter=delimiter,
        header=None if read_without_header else "infer",
        skiprows=skiprows,
        comment="#",
    )

    # Remove rows and columns where all values are NaN in the DataFrame:
    df = df.dropna(axis=0, how="all").dropna(axis=1, how="all")

    # Converting to NumPy array
    xydata = df.to_numpy(float)
    if xydata.size == 0:
        raise ValueError("Unable to read CSV file (no supported data after cleaning)")

    xlabel, ylabels, xunit, yunits = get_labels_units_from_dataframe(df)
    return xydata, xlabel, xunit, ylabels, yunits, header


def write_csv(
    filename: str,
    xydata: np.ndarray,
    xlabel: str | None,
    xunit: str | None,
    ylabels: list[str] | None,
    yunits: list[str] | None,
    header: str | None,
) -> None:
    """Write CSV data.

    Args:
        filename: CSV file name
        xydata: XY data
        xlabel: X label
        xunit: X unit
        ylabels: Y labels
        yunits: Y units
        header: Header
    """
    labels = ""
    delimiter = ","
    if len(ylabels) == 1:
        ylabels = ["Y"] if not ylabels[0] else ylabels
    elif ylabels:
        ylabels = [f"Y{i+1}" if not label else label for i, label in enumerate(ylabels)]
        if yunits:
            ylabels = [
                f"{label} ({unit})" if unit else label
                for label, unit in zip(ylabels, yunits)
            ]
    if ylabels:
        xlabel = xlabel or "X"
        if xunit:
            xlabel += f" ({xunit})"
        labels = delimiter.join([xlabel] + ylabels)
    df = pd.DataFrame(xydata.T, columns=[xlabel] + ylabels)
    df.to_csv(filename, index=False, header=labels, sep=delimiter)
    # Add header if present
    if header:
        with open(filename, "r+", encoding="utf-8") as file:
            content = file.read()
            file.seek(0, 0)
            file.write(header + "\n" + content)



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""Signal base line selection dialog."""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...

from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
from guidata.configtools import get_icon
from plotpy.builder import make
from plotpy.plot import PlotDialog

from cdl.config import _
from cdl.core.model.signal import CURVESTYLES

if TYPE_CHECKING:
    from plotpy.items import CurveItem, Marker, XRangeSelection
    from qtpy.QtWidgets import QWidget

    from cdl.obj import SignalObj


class SignalBaselineDialog(PlotDialog):
    """Signal baseline selection dialog.

    Args:
        signal: signal object
        parent: parent widget. Defaults to None.
    """

    def __init__(self, signal: SignalObj, parent: QWidget | None = None) -> None:
        self.__curve_styles = CURVESTYLES.style_generator()
        self.__baseline: float | None = None
        self.__x_range: tuple[float, float] = [np.nan, np.nan]
        self.curve: CurveItem | None = None
        self.cursor: Marker | None = None
        self.xrange: XRangeSelection | None = None
        super().__init__(title=_("Signal baseline selection"), edit=True, parent=parent)
        self.setObjectName("baselineselection")
        if parent is None:
            self.setWindowIcon(get_icon("DataLab.svg"))
        legend = make.legend("TR")
        self.get_plot().add_item(legend)
        self.__signal = signal.copy()
        self.__setup_dialog()

    def __setup_dialog(self) -> None:
        """Setup dialog box"""
        obj = self.__signal
        with CURVESTYLES.alternative(self.__curve_styles):
            self.curve = obj.make_item()
        plot = self.get_plot()
        plot.set_antialiasing(True)
        plot.SIG_RANGE_CHANGED.connect(self.xrange_changed)
        plot.SIG_MARKER_CHANGED.connect(self.cursor_changed)
        self.cursor = make.hcursor(0.0, _("Base line") + " = %g")
        self.xrange = make.range(obj.x[0], obj.x[int(0.2 * len(obj.x))])
        for item in (self.curve, self.cursor, self.xrange):
            plot.add_item(item)
        plot.replot()
        plot.set_active_item(self.xrange)
        self.xrange_changed(self.xrange, *self.xrange.get_range())

    # pylint: disable=unused-argument
    def xrange_changed(self, item: XRangeSelection, xmin: float, xmax: float) -> None:
        """X range changed"""
        self.__x_range = sorted([xmin, xmax])
        imin, imax = np.searchsorted(self.__signal.x, self.__x_range)
        if imin == imax:
            return
        self.cursor.set_pos(0, np.mean(self.__signal.y[imin:imax]))
        plot = self.get_plot()
        plot.replot()

    def cursor_changed(self, item: Marker) -> None:
        """Cursor changed"""
        _x, self.__baseline = item.get_pos()

    def get_baseline(self) -> float:
        """Get baseline"""
        return self.__baseline

    def get_x_range(self) -> tuple[float, float]:
        """Get x range"""
        return self.__x_range



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
Image tools test

Simple image dialog for testing all image tools available in DataLab
"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...
# guitest: show

from guidata.qthelpers import qt_app_context
from plotpy.builder import make

from cdl.tests.data import create_noisygauss_image
from cdl.utils.vistools import view_image_items


def test_image_tools_unit():
    """Image tools test"""
    with qt_app_context():
        data = create_noisygauss_image().data
        items = [make.image(data, interpolation="nearest", eliminate_outliers=2.0)]
        view_image_items(items)


if __name__ == "__main__":
    test_image_tools_unit()



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
ROI objects unit tests
"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...
# guitest: show

from __future__ import annotations

import numpy as np
from guidata.qthelpers import qt_app_context

from cdl.core.gui.roieditor import plot_item_to_single_roi
from cdl.env import execenv
from cdl.obj import (
    ImageObj,
    ImageROI,
    SignalObj,
    SignalROI,
    create_image_roi,
    create_signal_roi,
)
from cdl.tests.data import create_multigauss_image, create_paracetamol_signal

CLASS_NAME = "class_name"


def __conversion_methods(roi: SignalROI | ImageROI, obj: SignalObj | ImageObj) -> None:
    """Test conversion methods for single ROI objects"""
    execenv.print("    test `to_dict` and `from_dict` methods")
    roi_dict = roi.to_dict()
    roi_new = obj.get_roi_class().from_dict(roi_dict)
    assert roi.get_single_roi(0) == roi_new.get_single_roi(0)

    execenv.print("    test `to_plot_item` and `from_plot_item` methods: ", end="")
    single_roi = roi.get_single_roi(0)
    with qt_app_context(exec_loop=False):
        plot_item = single_roi.to_plot_item(obj)
        sroi_new = plot_item_to_single_roi(plot_item)
        orig_coords = [float(val) for val in single_roi.get_physical_coords(obj)]
        new_coords = [float(val) for val in sroi_new.get_physical_coords(obj)]
        execenv.print(f"{orig_coords} --> {new_coords}")
        assert np.array_equal(orig_coords, new_coords)


def test_create_signal_roi() -> None:
    """Test create_signal_roi"""

    # ROI coordinates: for each ROI type, the coordinates are given for indices=True
    # and indices=False (physical coordinates)
    coords = {
        "segment": {
            CLASS_NAME: "SegmentROI",
            True: [50, 100],  # indices [x0, dx]
            False: [7.5, 10.0],  # physical
        },
    }

    obj = create_paracetamol_signal()

    for indices in (True, False):
        execenv.print("indices:", indices)

        for geometry in coords:
            execenv.print("  geometry:", geometry)

            roi = create_signal_roi(coords[geometry][indices], indices=indices)

            sroi = roi.get_single_roi(0)
            assert sroi.__class__.__name__ == coords[geometry][CLASS_NAME]

            cds_ind = [int(val) for val in sroi.get_indices_coords(obj)]
            assert cds_ind == coords[geometry][True]

            cds_phys = [float(val) for val in sroi.get_physical_coords(obj)]
            assert cds_phys == coords[geometry][False]

            execenv.print("    get_physical_coords:", cds_phys)
            execenv.print("    get_indices_coords: ", cds_ind)

            __conversion_methods(roi, obj)


def test_create_image_roi() -> None:
    """Test create_image_roi"""

    # ROI coordinates: for each ROI type, the coordinates are given for indices=True
    # and indices=False (physical coordinates)
    coords = {
        "rectangle": {
            CLASS_NAME: "RectangularROI",
            True: [500, 750, 1000, 1250],  # indices [x0, y0, dx, dy]
            False: [500.5, 750.5, 1000.0, 1250.0],  # physical
        },
        "circle": {
            CLASS_NAME: "CircularROI",
            True: [1500, 1500, 500],  # indices [x0, y0, radius]
            False: [1500.5, 1500.5, 500.0],  # physical
        },
        "polygon": {
            CLASS_NAME: "PolygonalROI",
            True: [450, 150, 1300, 350, 1250, 950, 400, 1350],  # indices [x0, y0, ,...]
            False: [450.5, 150.5, 1300.5, 350.5, 1250.5, 950.5, 400.5, 1350.5],  # phys.
        },
    }

    obj = create_multigauss_image()

    for indices in (True, False):
        execenv.print("indices:", indices)

        for geometry in coords:
            execenv.print("  geometry:", geometry)

            roi = create_image_roi(geometry, coords[geometry][indices], indices=indices)

            sroi = roi.get_single_roi(0)
            assert sroi.__class__.__name__ == coords[geometry][CLASS_NAME]

            bbox_phys = [float(val) for val in sroi.get_bounding_box(obj)]
            if geometry in ("rectangle", "circle"):
                x0, y0, x1, y1 = obj.physical_to_indices(bbox_phys)
                if geometry == "rectangle":
                    coords_from_bbox = [int(xy) for xy in [x0, y0, x1 - x0, y1 - y0]]
                elif geometry == "circle":
                    coords_from_bbox = [
                        int(xy) for xy in [(x0 + x1) / 2, (y0 + y1) / 2, (x1 - x0) / 2]
                    ]
                assert coords_from_bbox == coords[geometry][True]

            cds_phys = [float(val) for val in sroi.get_physical_coords(obj)]
            assert cds_phys == coords[geometry][False]
            cds_ind = [int(val) for val in sroi.get_indices_coords(obj)]
            assert cds_ind == coords[geometry][True]

            execenv.print("    get_bounding_box:   ", bbox_phys)
            execenv.print("    get_physical_coords:", cds_phys)
            execenv.print("    get_indices_coords: ", cds_ind)

            __conversion_methods(roi, obj)


if __name__ == "__main__":
    test_create_signal_roi()
    test_create_image_roi()



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
.. Macro panel (see parent package :mod:`cdl.core.gui.panel`)
"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...

from __future__ import annotations

import os.path as osp
import re
from typing import TYPE_CHECKING

from guidata.config import CONF
from guidata.configtools import get_font, get_icon
from guidata.qthelpers import add_actions, create_action
from guidata.widgets.console.shell import PythonShellWidget
from guidata.widgets.dockable import DockableWidgetMixin
from qtpy import QtCore as QC
from qtpy import QtWidgets as QW
from qtpy.compat import getopenfilename, getsavefilename

from cdl.config import Conf, _
from cdl.core.gui.macroeditor import Macro
from cdl.core.gui.panel.base import AbstractPanel
from cdl.env import execenv
from cdl.utils.qthelpers import (
    create_menu_button,
    qt_try_loadsave_file,
    save_restore_stds,
)

if TYPE_CHECKING:
    from guidata.widgets.codeeditor import CodeEditor

    from cdl.core.io.native import NativeH5Reader, NativeH5Writer


class MacroTabs(QW.QTabWidget):
    """Macro tabwidget

    Args:
        parent (QWidget): Parent widget
    """

    SIG_CONTEXT_MENU = QC.Signal(QC.QPoint)
    SIG_RENAME = QC.Signal(int)
    SIG_REMOVE = QC.Signal(int)

    def __init__(self, parent=None) -> None:
        super().__init__(parent)
        self.setTabsClosable(True)
        self.setMovable(True)
        self.tabBarDoubleClicked.connect(self.__rename)
        self.tabCloseRequested.connect(self.__remove)
        self.__titles: list[str] = []

    def clear(self) -> None:
        """Override Qt method"""
        super().clear()
        self.__titles.clear()

    def contextMenuEvent(self, event):  # pylint: disable=C0103
        """Override Qt method"""
        self.SIG_CONTEXT_MENU.emit(event.globalPos())

    def __rename(self, index: int) -> None:
        """Rename tab

        Args:
            index: Index of the tab
        """
        self.SIG_RENAME.emit(index + 1)

    def __remove(self, index: int) -> None:
        """Remove tab

        Args:
            index: Index of the tab
        """
        self.SIG_REMOVE.emit(index + 1)

    def __update_tab_titles(self) -> None:
        """Update tab titles"""
        for number, title in enumerate(self.__titles, 1):
            self.setTabText(number - 1, f"{number:02d}: {title}")

    def add_tab(self, macro: Macro) -> int:
        """Add tab

        Args:
            macro: Macro object

        Returns:
            int: Number of the tab (starting at 1)
        """
        self.__titles.append(macro.title)
        index = self.addTab(macro.editor, "")
        self.__update_tab_titles()
        return index + 1  # Numbering starts at 1

    def remove_tab(self, number: int) -> None:
        """Remove tab

        Args:
            number: Number of the tab (starting at 1)
        """
        self.removeTab(number - 1)
        self.__titles.pop(number - 1)
        self.__update_tab_titles()

    def get_widget(self, number: int) -> CodeEditor:
        """Return macro editor widget at number

        Args:
            number: Number of the tab (starting at 1)

        Returns:
            Macro editor widget
        """
        return self.widget(number - 1)

    def set_current_number(self, number: int) -> None:
        """Set current tab number

        Args:
            number: Number of the tab (starting at 1)
        """
        self.setCurrentIndex(number - 1)

    def get_current_number(self) -> int:
        """Return current tab number

        Returns:
            int: Number of the tab (starting at 1)
        """
        return self.currentIndex() + 1

    def set_tab_title(self, number: int, name: str) -> None:
        """Set tab title

        Args:
            number: Number of the tab (starting at 1)
            name: Macro name
        """
        self.__titles[number - 1] = name
        self.__update_tab_titles()


class MacroPanel(AbstractPanel, DockableWidgetMixin):
    """Macro Panel widget

    Args:
        parent (QWidget): Parent widget
    """

    LOCATION = QC.Qt.RightDockWidgetArea
    PANEL_STR = _("Macro panel")

    H5_PREFIX = "DataLab_Mac"

    SIG_OBJECT_MODIFIED = QC.Signal()

    FILE_FILTERS = f"{_('Python files')} (*.py)"

    def __init__(self, parent: QW.QWidget | None = None) -> None:
        super().__init__(parent)
        self.setWindowTitle(self.PANEL_STR)
        self.setWindowIcon(get_icon("libre-gui-cogs.svg"))
        self.setOrientation(QC.Qt.Vertical)

        self.context_menu = QW.QMenu()
        self.tabwidget_tb = QW.QToolBar(self)
        self.tabwidget_tb.setOrientation(QC.Qt.Vertical)

        self.console = PythonShellWidget(self, read_only=True)
        self.console.setMaximumBlockCount(5000)
        font = get_font(CONF, "console")
        font.setPointSize(10)
        self.console.set_font(font)
        self.console.write(_("-***- Macro Console -***-"), prompt=True)

        self.tabwidget = MacroTabs(self)
        self.tabwidget.SIG_RENAME.connect(self.rename_macro)
        self.tabwidget.SIG_REMOVE.connect(self.remove_macro)
        self.tabwidget.currentChanged.connect(self.__update_actions)

        tabwidget_with_tb = QW.QWidget(self)
        tabwidget_with_tb.setLayout(QW.QHBoxLayout())
        tabwidget_with_tb.layout().addWidget(self.tabwidget_tb)
        tabwidget_with_tb.layout().addWidget(self.tabwidget)

        # Put console in a groupbox to have a title
        console_groupbox = QW.QGroupBox(_("Console"), self)
        console_groupbox.setLayout(QW.QHBoxLayout())
        console_groupbox.layout().addWidget(self.console)
        # Put console groupbox in a frame to have a nice margin
        console_frame = QW.QFrame(self)
        console_frame.setLayout(QW.QHBoxLayout())
        console_frame.layout().addWidget(console_groupbox)

        for widget in (tabwidget_with_tb, console_frame):
            self.addWidget(widget)
        self.setStretchFactor(0, 2)
        self.setStretchFactor(1, 1)

        self.run_action = None
        self.stop_action = None
        self.obj_actions: list[QW.QAction] = []  # Object-dependent actions
        self.__macros: list[Macro] = []

        self.setup_actions()

    def update_color_mode(self) -> None:
        """Update color mode according to the current theme"""
        self.console.update_color_mode()
        for macro in self.__macros:
            macro.editor.update_color_mode()

    # ------AbstractPanel interface-----------------------------------------------------
    # pylint: disable=unused-argument
    def get_serializable_name(self, obj: Macro) -> str:
        """Return serializable name of object"""
        title = re.sub("[^-a-zA-Z0-9_.() ]+", "", obj.title.replace("/", "_"))
        name = f"{obj.PREFIX}{(self.__macros.index(obj)+1):03d}: {title}"
        return name

    def serialize_to_hdf5(self, writer: NativeH5Writer) -> None:
        """Serialize whole panel to a HDF5 file

        Args:
            writer: HDF5 writer
        """
        with writer.group(self.H5_PREFIX):
            for obj in self.__macros:
                self.serialize_object_to_hdf5(obj, writer)

    def deserialize_from_hdf5(self, reader: NativeH5Reader) -> None:
        """Deserialize whole panel from a HDF5 file

        Args:
            reader: HDF5 reader
        """
        with reader.group(self.H5_PREFIX):
            for name in reader.h5.get(self.H5_PREFIX, []):
                #  Contrary to signal or image panels, macros are not stored
                #  in a group but directly in the root of the HDF5 file
                obj = self.deserialize_object_from_hdf5(reader, name)
                self.add_object(obj)

    def __len__(self) -> int:
        """Return number of objects"""
        return len(self.__macros)

    def __getitem__(self, nb: int) -> Macro:
        """Return object from its number (1 to N)"""
        return self.__macros[nb - 1]

    def __iter__(self):
        """Iterate over objects"""
        return iter(self.__macros)

    def create_object(self) -> Macro:
        """Create object.

        Returns:
            Macro object
        """
        macro = Macro(self.console)
        macro.objectNameChanged.connect(self.macro_name_changed)
        macro.STARTED.connect(
            lambda orig_macro=macro: self.macro_state_changed(orig_macro, True)
        )
        macro.FINISHED.connect(
            lambda orig_macro=macro: self.macro_state_changed(orig_macro, False)
        )
        macro.MODIFIED.connect(self.macro_contents_changed)
        return macro

    def add_object(self, obj: Macro) -> None:
        """Add object.

        Args:
            obj: Macro object
        """
        self.__macros.append(obj)
        number = self.tabwidget.add_tab(obj)
        self.SIG_OBJECT_ADDED.emit()
        self.tabwidget.set_current_number(number)

    def remove_all_objects(self) -> None:
        """Remove all objects"""
        self.tabwidget.clear()
        self.__macros.clear()
        super().remove_all_objects()

    # ---- Macro panel API -------------------------------------------------------------
    def setup_actions(self) -> None:
        """Setup macro menu actions"""
        self.run_action = create_action(
            self,
            _("Run macro"),
            icon=get_icon("libre-camera-flash-on.svg"),
            triggered=self.run_macro,
            shortcut="Ctrl+F5",
        )
        self.stop_action = create_action(
            self,
            _("Stop macro"),
            icon=get_icon("libre-camera-flash-off.svg"),
            triggered=self.stop_macro,
            shortcut="Shift+F5",
        )
        self.stop_action.setDisabled(True)
        add_act = create_action(
            self,
            _("New macro"),
            icon=get_icon("libre-gui-add.svg"),
            triggered=self.add_macro,
        )
        ren_act = create_action(
            self,
            _("Rename macro"),
            icon=get_icon("libre-gui-pencil.svg"),
            triggered=self.rename_macro,
        )
        imp_act = create_action(
            self,
            _("Import macro from file"),
            icon=get_icon("fileopen_py.svg"),
            triggered=self.import_macro_from_file,
        )
        exp_act = create_action(
            self,
            _("Export macro to file"),
            icon=get_icon("filesave_py.svg"),
            triggered=self.export_macro_to_file,
        )
        rem_act = create_action(
            self,
            _("Remove macro"),
            icon=get_icon("libre-gui-action-delete.svg"),
            triggered=self.remove_macro,
        )
        self.obj_actions += [
            self.run_action,
            self.stop_action,
            ren_act,
            exp_act,
            rem_act,
        ]

        self.tabwidget.SIG_CONTEXT_MENU.connect(self.context_menu.popup)

        tabwidget_corner = QW.QToolBar(_("Macro editor toolbar"), self)
        self.context_menu.aboutToShow.connect(self.__update_actions)
        tabwidget_menu_btn = create_menu_button(self, self.context_menu)
        tabwidget_corner.addWidget(tabwidget_menu_btn)
        self.tabwidget.setCornerWidget(tabwidget_corner)

        main_actions = [
            self.run_action,
            self.stop_action,
            None,
            add_act,
            ren_act,
            imp_act,
            exp_act,
        ]
        add_actions(tabwidget_corner, [self.run_action, self.stop_action])
        add_actions(self.tabwidget_tb, main_actions)
        add_actions(self.context_menu, main_actions + [None, rem_act])

        self.__update_actions()

    def __update_actions(self) -> None:
        """Update actions"""
        not_empty = self.tabwidget.count() > 0
        for action in self.obj_actions:
            action.setEnabled(not_empty)
        if not_empty:
            macro = self.get_macro()
            if macro is not None:
                macro: Macro
                self.macro_state_changed(macro, macro.is_running())

    def get_macro(self, number_or_title: int | str | None = None) -> Macro | None:
        """Return macro at number (if number is None, return current macro)

        Args:
            number: Number of the macro (starting at 1) or title of the macro.
             Defaults to None (current macro).

        Returns:
            Macro object or None (if not found)
        """
        if number_or_title is None:
            number_or_title = self.tabwidget.get_current_number()
        if isinstance(number_or_title, str):
            return self.get_macro(self.get_number_from_title(number_or_title))
        for macro in self.__macros:
            if self.tabwidget.get_widget(number_or_title) is macro.editor:
                return macro
        return None

    def get_number_from_title(self, title: str) -> int | None:
        """Return macro number from title

        Args:
            title: Title of the macro

        Returns:
            Number of the macro (starting at 1) or None (if not found)
        """
        for number in range(1, self.tabwidget.count() + 1):
            if self.tabwidget.tabText(number - 1).endswith(title):
                return number
        return None

    def get_number_from_macro(self, macro: Macro) -> int | None:
        """Return macro number from macro object

        Args:
            macro: Macro object

        Returns:
            Number of the macro (starting at 1) or None (if not found)
        """
        for number in range(1, self.tabwidget.count() + 1):
            if self.tabwidget.get_widget(number) is macro.editor:
                return number
        return None

    def get_macro_titles(self) -> list[str]:
        """Return list of macro titles"""
        return [macro.title for macro in self.__macros]

    def macro_contents_changed(self) -> None:
        """One of the macro contents has changed"""
        self.SIG_OBJECT_MODIFIED.emit()

    def run_macro(self, number_or_title: int | str | None = None) -> None:
        """Run current macro

        Args:
            number: Number of the macro (starting at 1). Defaults to None (run
             current macro, or does nothing if there is no macro).
        """
        macro = self.get_macro(number_or_title)
        if macro is not None:
            macro: Macro
            macro.run()

    def stop_macro(self, number_or_title: int | str | None = None) -> None:
        """Stop current macro

        Args:
            number: Number of the macro (starting at 1). Defaults to None (run
             current macro, or does nothing if there is no macro).
        """
        macro = self.get_macro(number_or_title)
        if macro is not None:
            macro: Macro
            macro.kill()

    def macro_state_changed(self, orig_macro: Macro, state: bool) -> None:
        """Macro state has changed (True: started, False: stopped)

        Args:
            orig_macro: Macro object
            state: State of the macro
        """
        macro = self.get_macro()
        if macro is orig_macro:
            self.run_action.setEnabled(not state)
            self.stop_action.setEnabled(state)

    def add_macro(self) -> Macro:
        """Add macro, optionally with name

        Returns:
            Macro object
        """
        macro = self.create_object()
        self.add_object(macro)
        if not macro.title:
            self.rename_macro()
        return macro

    def macro_name_changed(self, name: str) -> None:
        """Macro name has been changed

        Args:
            name: New name of the macro
        """
        number = self.get_number_from_macro(self.sender())
        if number is not None:
            self.tabwidget.set_tab_title(number, name)

    def rename_macro(self, number: int | None = None, title: str | None = None) -> None:
        """Rename macro

        Args:
            number: Number of the macro (starting at 1). Defaults to None.
            title: Title of the macro. Defaults to None.
        """
        macro = self.get_macro(number)
        assert isinstance(macro, Macro)
        if title is None:
            title, valid = QW.QInputDialog.getText(
                self,
                _("Rename"),
                _("New title:"),
                QW.QLineEdit.Normal,
                macro.title,
            )
            title = title if valid else None
        if title:
            macro.title = title
            if number is not None:
                self.tabwidget.set_current_number(number)

    def export_macro_to_file(
        self, number_or_title: int | str | None = None, filename: str | None = None
    ) -> None:
        """Export macro to file

        Args:
            number_or_title: Number of the macro (starting at 1) or title of the macro.
             Defaults to None.
            filename: Filename. Defaults to None.

        Raises:
            ValueError: If title is not found
        """
        macro = self.get_macro(number_or_title)
        assert isinstance(macro, Macro)
        if filename is None:
            basedir = Conf.main.base_dir.get()
            with save_restore_stds():
                filename, _filt = getsavefilename(
                    self, _("Save as"), basedir, self.FILE_FILTERS
                )
        if filename:
            with qt_try_loadsave_file(self.parent(), filename, "save"):
                Conf.main.base_dir.set(filename)
                macro.title = osp.basename(filename)
                macro.to_file(filename)

    def import_macro_from_file(self, filename: str | None = None) -> int:
        """Import macro from file

        Args:
            filename: Filename. Defaults to None.

        Returns:
            Number of the macro (starting at 1)
        """
        if filename is None:
            basedir = Conf.main.base_dir.get()
            with save_restore_stds():
                filename, _filt = getopenfilename(
                    self, _("Open"), basedir, self.FILE_FILTERS
                )
        if filename:
            with qt_try_loadsave_file(self.parent(), filename, "load"):
                Conf.main.base_dir.set(filename)
                macro = self.add_macro()
                macro.from_file(filename)
            return self.get_number_from_macro(macro)
        return -1

    def remove_macro(self, number_or_title: int | str | None = None) -> None:
        """Remove macro

        Args:
            number_or_title: Number of the macro (starting at 1) or title of the macro.
             Defaults to None.
        """
        if number_or_title is None:
            number_or_title = self.tabwidget.get_current_number()
        if isinstance(number_or_title, str):
            number_or_title = self.get_number_from_title(number_or_title)
        txt = "<br>".join(
            [
                _(
                    "When closed, the macro is <u>permanently destroyed</u>, "
                    "unless it has been exported first."
                ),
                "",
                _("Do you want to continue?"),
            ]
        )
        btns = QW.QMessageBox.StandardButton.Yes | QW.QMessageBox.StandardButton.No
        if execenv.unattended:
            choice = QW.QMessageBox.StandardButton.Yes
        else:
            choice = QW.QMessageBox.warning(self, self.windowTitle(), txt, btns)
        if choice == QW.QMessageBox.StandardButton.Yes:
            self.tabwidget.remove_tab(number_or_title)
            self.__macros.pop(number_or_title - 1)
            self.SIG_OBJECT_REMOVED.emit()



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
Signal FFT unit test.
"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...
# pylint: disable=duplicate-code
# guitest: show

from __future__ import annotations

import numpy as np
import pytest
import scipy.signal as sps
from guidata.qthelpers import qt_app_context

import cdl.algorithms.signal as alg
import cdl.computation.signal as cps
import cdl.obj
import cdl.param
import cdl.tests.data as ctd
from cdl.env import execenv
from cdl.utils.tests import check_array_result, check_scalar_result
from cdl.utils.vistools import view_curves


def test_signal_fft_interactive() -> None:
    """1D FFT interactive test."""
    with qt_app_context():
        newparam = cdl.obj.new_signal_param(stype=cdl.obj.SignalTypes.COSINUS, size=500)

        # *** Note ***
        #
        # We set xmin to 0.0 to be able to compare the X data of the original and
        # reconstructed signals, because the FFT do not preserve the X data (phase is
        # lost, sampling rate is assumed to be constant), so that comparing the X data
        # is not meaningful if xmin is different.
        newparam.xmin = 0.0

        s1 = cdl.obj.create_signal_from_param(newparam)
        t, y = s1.xydata
        f, s = alg.fft1d(t, y)
        t2, y2 = alg.ifft1d(f, s)
        execenv.print("Comparing original and FFT/iFFT signals...", end=" ")
        execenv.print("OK")
        check_array_result("Signal FFT/iFFT X data", t2, t)
        check_array_result("Signal FFT/iFFT Y data", y2, y)
        view_curves([(t, y), (t2, y2)])


@pytest.mark.validation
def test_signal_fft() -> None:
    """1D FFT validation test."""
    freq = 50.0
    size = 10000

    # See note in the interactive test above
    xmin = 0.0

    s1 = ctd.create_periodic_signal(
        cdl.obj.SignalTypes.COSINUS, freq=freq, size=size, xmin=xmin
    )
    fft = cps.compute_fft(s1)
    ifft = cps.compute_ifft(fft)

    # Check that the inverse FFT reconstructs the original signal
    check_array_result("Cosine signal FFT/iFFT X reconstruction", s1.y, ifft.y.real)
    check_array_result("Cosine signal FFT/iFFT Y reconstruction", s1.x, ifft.x.real)

    # Check FFT properties
    mag = np.abs(fft.y)

    # Find the peak in the FFT
    ipk1, ipk2 = np.argmax(mag[: size // 2]), np.argmax(mag[size // 2 :]) + size // 2
    fpk1, fpk2 = fft.x[ipk1], fft.x[ipk2]

    # Verify the peak frequencies are correct
    check_scalar_result("Cosine negative frequency", fpk1, -freq, rtol=0.001)
    check_scalar_result("Cosine positive frequency", fpk2, freq, rtol=0.001)

    # Verify the magnitude at the peak
    exp_mag = size / 2
    check_scalar_result("Cosine peak magnitude", mag[ipk1], exp_mag, rtol=0.05)

    # Verify the symmetry of the FFT
    check_array_result(
        "FFT symmetry",
        mag[1 : size // 2],
        mag[1 + size // 2 :][::-1],
    )


@pytest.mark.skip(reason="Already covered by the `test_signal_fft` test.")
@pytest.mark.validation
def test_signal_ifft() -> None:
    """1D iFFT validation test."""
    # This is just a way of marking the iFFT test as a validation test because it is
    # already covered by the FFT test above (there is no need to repeat the same test).


@pytest.mark.validation
def test_signal_magnitude_spectrum() -> None:
    """1D magnitude spectrum validation test."""
    freq = 50.0
    size = 10000

    s1 = ctd.create_periodic_signal(cdl.obj.SignalTypes.COSINUS, freq=freq, size=size)
    fft = cps.compute_fft(s1)
    mag = cps.compute_magnitude_spectrum(s1)
    fpk1 = fft.x[np.argmax(mag.y[: size // 2])]
    check_scalar_result("Cosine negative frequency", fpk1, -freq, rtol=0.001)

    # Check that the magnitude spectrum is correct
    check_array_result("Cosine signal magnitude spectrum X", mag.x, fft.x.real)
    check_array_result("Cosine signal magnitude spectrum Y", mag.y, np.abs(fft.y))


@pytest.mark.validation
def test_signal_phase_spectrum() -> None:
    """1D phase spectrum validation test."""
    freq = 50.0
    size = 10000

    s1 = ctd.create_periodic_signal(cdl.obj.SignalTypes.COSINUS, freq=freq, size=size)
    fft = cps.compute_fft(s1)
    phase = cps.compute_phase_spectrum(s1)
    fpk1 = fft.x[np.argmax(phase.y[: size // 2])]
    check_scalar_result("Cosine negative frequency", fpk1, -freq, rtol=0.001)

    # Check that the phase spectrum is correct
    check_array_result("Cosine signal phase spectrum X", phase.x, fft.x.real)
    exp_phase = np.rad2deg(np.angle(fft.y))
    check_array_result("Cosine signal phase spectrum Y", phase.y, exp_phase)


@pytest.mark.validation
def test_signal_psd() -> None:
    """1D Power Spectral Density validation test."""
    freq = 50.0
    size = 10000

    s1 = ctd.create_periodic_signal(cdl.obj.SignalTypes.COSINUS, freq=freq, size=size)
    param = cdl.param.SpectrumParam()
    for log_scale in (False, True):
        param.log = log_scale
        psd = cps.compute_psd(s1, param)

        # Check that the PSD is correct (Welch's method is used by default)
        exp_x, exp_y = sps.welch(s1.y, fs=1.0 / (s1.x[1] - s1.x[0]))

        if log_scale:
            exp_y = 10 * np.log10(exp_y)

        check_array_result(f"Cosine signal PSD X (log={log_scale})", psd.x, exp_x)
        check_array_result(f"Cosine signal PSD Y (log={log_scale})", psd.y, exp_y)


if __name__ == "__main__":
    test_signal_fft_interactive()
    test_signal_fft()
    test_signal_magnitude_spectrum()
    test_signal_phase_spectrum()
    test_signal_psd()



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
Memory leak test

DataLab application memory leak test.
"""

# guitest: skip

import os

import numpy as np
import psutil
from guidata.qthelpers import qt_app_context

from cdl.env import execenv
from cdl.tests.features.control.embedded1_unit_test import HostWindow
from cdl.utils.vistools import view_curves


def memory_leak_test(iterations=100):
    """Test for memory leak"""
    with qt_app_context():
        proc = psutil.Process(os.getpid())
        mainview = HostWindow()
        mainview.show()
        memlist = []
        for i in range(iterations):
            mainview.init_cdl()
            mainview.close_cdl()
            memdata = proc.memory_info().vms / 1024**2
            memlist.append(memdata)
            execenv.print(i + 1, ":", memdata, "MB")
        view_curves(
            np.array(memlist),
            title="Memory leak test for DataLab application",
            ylabel="Memory (MB)",
        )


if __name__ == "__main__":
    memory_leak_test()



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
Morphology computation module
-----------------------------

"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...

# Note:
# ----
# All dataset classes must also be imported in the cdl.computation.param module.

from __future__ import annotations

import guidata.dataset as gds
from skimage import morphology

from cdl.computation.image import dst_11, restore_data_outside_roi
from cdl.config import _
from cdl.obj import ImageObj


class MorphologyParam(gds.DataSet):
    """White Top-Hat parameters"""

    radius = gds.IntItem(
        _("Radius"), default=1, min=1, help=_("Footprint (disk) radius.")
    )


def compute_white_tophat(src: ImageObj, p: MorphologyParam) -> ImageObj:
    """Compute White Top-Hat with :py:func:`skimage.morphology.white_tophat`

    Args:
        src: input image object
        p: parameters

    Returns:
        Output image object
    """
    dst = dst_11(src, "white_tophat", f"radius={p.radius}")
    dst.data = morphology.white_tophat(src.data, morphology.disk(p.radius))
    restore_data_outside_roi(dst, src)
    return dst


def compute_black_tophat(src: ImageObj, p: MorphologyParam) -> ImageObj:
    """Compute Black Top-Hat with :py:func:`skimage.morphology.black_tophat`

    Args:
        src: input image object
        p: parameters

    Returns:
        Output image object
    """
    dst = dst_11(src, "black_tophat", f"radius={p.radius}")
    dst.data = morphology.black_tophat(src.data, morphology.disk(p.radius))
    restore_data_outside_roi(dst, src)
    return dst


def compute_erosion(src: ImageObj, p: MorphologyParam) -> ImageObj:
    """Compute Erosion with :py:func:`skimage.morphology.erosion`

    Args:
        src: input image object
        p: parameters

    Returns:
        Output image object
    """
    dst = dst_11(src, "erosion", f"radius={p.radius}")
    dst.data = morphology.erosion(src.data, morphology.disk(p.radius))
    restore_data_outside_roi(dst, src)
    return dst


def compute_dilation(src: ImageObj, p: MorphologyParam) -> ImageObj:
    """Compute Dilation with :py:func:`skimage.morphology.dilation`

    Args:
        src: input image object
        p: parameters

    Returns:
        Output image object
    """
    dst = dst_11(src, "dilation", f"radius={p.radius}")
    dst.data = morphology.dilation(src.data, morphology.disk(p.radius))
    restore_data_outside_roi(dst, src)
    return dst


def compute_opening(src: ImageObj, p: MorphologyParam) -> ImageObj:
    """Compute morphological opening with :py:func:`skimage.morphology.opening`

    Args:
        src: input image object
        p: parameters

    Returns:
        Output image object
    """
    dst = dst_11(src, "opening", f"radius={p.radius}")
    dst.data = morphology.opening(src.data, morphology.disk(p.radius))
    restore_data_outside_roi(dst, src)
    return dst


def compute_closing(src: ImageObj, p: MorphologyParam) -> ImageObj:
    """Compute morphological closing with :py:func:`skimage.morphology.closing`

    Args:
        src: input image object
        p: parameters

    Returns:
        Output image object
    """
    dst = dst_11(src, "closing", f"radius={p.radius}")
    dst.data = morphology.closing(src.data, morphology.disk(p.radius))
    restore_data_outside_roi(dst, src)
    return dst



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
Macro-commands test scenario
----------------------------

Testing all the macro-commands features.
"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...
# guitest: show

import os.path as osp
import time

from cdl.core.gui.macroeditor import Macro
from cdl.core.gui.main import CDLMainWindow
from cdl.tests import cdltest_app_context
from cdl.utils.tests import CDLTemporaryDirectory


def add_macro_sample(win: CDLMainWindow, index: int) -> Macro:
    """Add a macro sample to the macro panel

    Args:
        win: CDLMainWindow
        index: index of the macro sample to add
    """
    macro = win.macropanel.add_macro()
    macro.title = f"Test macro {index}"
    samples = [
        "import time; [print(f'{i}:{time.sleep(.1)}') for i in range(50)]",
        "print('Hello world')",
    ]
    assert index < len(samples), f"index={index} is out of range"
    macro.set_code(samples[index])
    return macro


def test_scenario_macro() -> None:
    """Example of high-level test scenario with HDF5 file"""
    with CDLTemporaryDirectory() as tmpdir:
        with cdltest_app_context(console=False) as win:
            win.set_current_panel("macro")
            add_macro_sample(win, 0)
            win.macropanel.run_macro()
            time.sleep(1)
            win.macropanel.stop_macro()
            code2 = add_macro_sample(win, 1).get_code()
            fname = osp.join(tmpdir, "test.macro")
            win.macropanel.export_macro_to_file(2, fname)
            win.macropanel.remove_macro()
            macro_nb = win.macropanel.import_macro_from_file(fname)
            macro2 = win.macropanel.get_macro(macro_nb)
            assert macro2.get_code() == code2, "Macro code is not the same"


if __name__ == "__main__":
    test_scenario_macro()



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""Image background selection dialog."""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...

from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
from guidata.configtools import get_icon
from plotpy.builder import make
from plotpy.plot import PlotDialog

from cdl.config import _

if TYPE_CHECKING:
    from plotpy.items import ImageItem, RangeComputation2d, RectangleShape
    from qtpy.QtWidgets import QWidget

    from cdl.obj import ImageObj


class ImageBackgroundDialog(PlotDialog):
    """Image background selection dialog.

    Args:
        image: image object
        parent: parent widget. Defaults to None.
    """

    def __init__(self, image: ImageObj, parent: QWidget | None = None) -> None:
        self.__background: float | None = None
        self.__rect_coords: tuple[float, float, float, float] | None = None
        self.imageitem: ImageItem | None = None
        self.rectarea: RectangleShape | None = None
        self.comput2d: RangeComputation2d | None = None
        super().__init__(
            title=_("Image background selection"), edit=True, parent=parent
        )
        self.setObjectName("backgroundselection")
        if parent is None:
            self.setWindowIcon(get_icon("DataLab.svg"))
        self.__image = image.copy()
        self.__setup_dialog()

    def __compute_background(
        self,
        x: np.ndarray,  # pylint: disable=unused-argument
        y: np.ndarray,  # pylint: disable=unused-argument
        z: np.ndarray,
    ) -> float:
        """Compute background value"""
        self.__rect_coords = self.rectarea.get_rect()
        self.__background = z.mean()
        return self.__background

    def __setup_dialog(self) -> None:
        """Setup dialog box"""
        obj = self.__image
        self.imageitem = obj.make_item()
        plot = self.get_plot()
        self.rectarea = make.rectangle(
            obj.x0, obj.y0, obj.xc + obj.dx, obj.yc + obj.dy, _("Background area")
        )
        self.comput2d = make.computation2d(
            self.rectarea,
            "TL",
            _("Background value:") + " %g",
            self.imageitem,
            self.__compute_background,
        )
        for item in (self.imageitem, self.rectarea, self.comput2d):
            plot.add_item(item)
        plot.replot()
        plot.set_active_item(self.rectarea)

    def get_background(self) -> float:
        """Get background value"""
        return self.__background

    def get_rect_coords(self) -> tuple[float, float, float, float]:
        """Get rectangle coordinates"""
        return self.__rect_coords



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
Unit tests for signal operations
--------------------------------

Features from the "Operations" menu are covered by this test.
The "Operations" menu contains basic operations on signals, such as
addition, multiplication, division, and more.
"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...

from __future__ import annotations

import numpy as np
import pytest

import cdl.computation.signal as cps
import cdl.obj
import cdl.param
import cdl.tests.data as ctd
from cdl.utils.tests import check_array_result


def __create_two_signals() -> tuple[cdl.obj.SignalObj, cdl.obj.SignalObj]:
    """Create two signals for testing."""
    s1 = ctd.create_periodic_signal(cdl.obj.SignalTypes.COSINUS, freq=50.0, size=100)
    s2 = ctd.create_periodic_signal(cdl.obj.SignalTypes.SINUS, freq=25.0, size=100)
    return s1, s2


def __create_one_signal_and_constant() -> (
    tuple[cdl.obj.SignalObj, cdl.param.ConstantParam]
):
    """Create one signal and a constant for testing."""
    s1 = ctd.create_periodic_signal(cdl.obj.SignalTypes.COSINUS, freq=50.0, size=100)
    param = cdl.param.ConstantParam.create(value=-np.pi)
    return s1, param


@pytest.mark.validation
def test_signal_addition() -> None:
    """Signal addition test."""
    s1, s2 = __create_two_signals()
    exp = s1.y + s2.y
    cps.compute_addition(s1, s2)
    res = s1.y
    check_array_result("Signal addition", res, exp)


@pytest.mark.validation
def test_signal_product() -> None:
    """Signal multiplication test."""
    s1, s2 = __create_two_signals()
    exp = s1.y * s2.y
    cps.compute_product(s1, s2)
    res = s1.y
    check_array_result("Signal multiplication", res, exp)


@pytest.mark.validation
def test_signal_difference() -> None:
    """Signal difference test."""
    s1, s2 = __create_two_signals()
    s3 = cps.compute_difference(s1, s2)
    check_array_result("Signal difference", s3.y, s1.y - s2.y)


@pytest.mark.validation
def test_signal_quadratic_difference() -> None:
    """Signal quadratic difference validation test."""
    s1, s2 = __create_two_signals()
    s3 = cps.compute_quadratic_difference(s1, s2)
    check_array_result("Signal quadratic difference", s3.y, (s1.y - s2.y) / np.sqrt(2))


@pytest.mark.validation
def test_signal_division() -> None:
    """Signal division test."""
    s1, s2 = __create_two_signals()
    s3 = cps.compute_division(s1, s2)
    check_array_result("Signal division", s3.y, s1.y / s2.y)


@pytest.mark.validation
def test_signal_addition_constant() -> None:
    """Signal addition with constant test."""
    s1, param = __create_one_signal_and_constant()
    s2 = cps.compute_addition_constant(s1, param)
    check_array_result("Signal addition with constant", s2.y, s1.y + param.value)


@pytest.mark.validation
def test_signal_product_constant() -> None:
    """Signal multiplication by constant test."""
    s1, param = __create_one_signal_and_constant()
    s2 = cps.compute_product_constant(s1, param)
    check_array_result("Signal multiplication by constant", s2.y, s1.y * param.value)


@pytest.mark.validation
def test_signal_difference_constant() -> None:
    """Signal difference with constant test."""
    s1, param = __create_one_signal_and_constant()
    s2 = cps.compute_difference_constant(s1, param)
    check_array_result("Signal difference with constant", s2.y, s1.y - param.value)


@pytest.mark.validation
def test_signal_division_constant() -> None:
    """Signal division by constant test."""
    s1, param = __create_one_signal_and_constant()
    s2 = cps.compute_division_constant(s1, param)
    check_array_result("Signal division by constant", s2.y, s1.y / param.value)


@pytest.mark.validation
def test_signal_abs() -> None:
    """Absolute value validation test."""
    s1 = __create_two_signals()[0]
    abs_signal = cps.compute_abs(s1)
    check_array_result("Absolute value", abs_signal.y, np.abs(s1.y))


@pytest.mark.validation
def test_signal_re() -> None:
    """Real part validation test."""
    s1 = __create_two_signals()[0]
    re_signal = cps.compute_re(s1)
    check_array_result("Real part", re_signal.y, np.real(s1.y))


@pytest.mark.validation
def test_signal_im() -> None:
    """Imaginary part validation test."""
    s1 = __create_two_signals()[0]
    im_signal = cps.compute_im(s1)
    check_array_result("Imaginary part", im_signal.y, np.imag(s1.y))


@pytest.mark.validation
def test_signal_astype() -> None:
    """Data type conversion validation test."""
    s1 = __create_two_signals()[0]
    for dtype_str in cps.VALID_DTYPES_STRLIST:
        p = cdl.param.DataTypeSParam.create(dtype_str=dtype_str)
        astype_signal = cps.compute_astype(s1, p)
        assert astype_signal.y.dtype == np.dtype(dtype_str)


@pytest.mark.validation
def test_signal_exp() -> None:
    """Exponential validation test."""
    s1 = __create_two_signals()[0]
    exp_signal = cps.compute_exp(s1)
    check_array_result("Exponential", exp_signal.y, np.exp(s1.y))


@pytest.mark.validation
def test_signal_log10() -> None:
    """Logarithm base 10 validation test."""
    s1 = __create_two_signals()[0]
    log10_signal = cps.compute_log10(cps.compute_exp(s1))
    check_array_result("Logarithm base 10", log10_signal.y, np.log10(np.exp(s1.y)))


@pytest.mark.validation
def test_signal_sqrt() -> None:
    """Square root validation test."""
    s1 = ctd.get_test_signal("paracetamol.txt")
    sqrt_signal = cps.compute_sqrt(s1)
    check_array_result("Square root", sqrt_signal.y, np.sqrt(s1.y))


@pytest.mark.validation
def test_signal_power() -> None:
    """Power validation test."""
    s1 = ctd.get_test_signal("paracetamol.txt")
    p = cdl.param.PowerParam.create(power=2.0)
    power_signal = cps.compute_power(s1, p)
    check_array_result("Power", power_signal.y, s1.y**p.power)


@pytest.mark.validation
def test_signal_arithmetic() -> None:
    """Arithmetic operations validation test."""
    s1, s2 = __create_two_signals()
    p = cdl.param.ArithmeticParam.create()
    for operator in p.operators:
        p.operator = operator
        for factor in (0.0, 1.0, 2.0):
            p.factor = factor
            for constant in (0.0, 1.0, 2.0):
                p.constant = constant
                s3 = cps.compute_arithmetic(s1, s2, p)
                if operator == "+":
                    exp = s1.y + s2.y
                elif operator == "×":
                    exp = s1.y * s2.y
                elif operator == "-":
                    exp = s1.y - s2.y
                elif operator == "/":
                    exp = s1.y / s2.y
                exp = exp * factor + constant
                check_array_result(f"Arithmetic [{p.get_operation()}]", s3.y, exp)


if __name__ == "__main__":
    test_signal_addition()
    test_signal_product()
    test_signal_difference()
    test_signal_quadratic_difference()
    test_signal_division()
    test_signal_addition_constant()
    test_signal_product_constant()
    test_signal_difference_constant()
    test_signal_division_constant()
    test_signal_abs()
    test_signal_re()
    test_signal_im()
    test_signal_astype()
    test_signal_exp()
    test_signal_log10()
    test_signal_sqrt()
    test_signal_power()
    test_signal_arithmetic()



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
DataLab mainwindow automation test
----------------------------------

This test scenario shows how to use the DataLab application in a test
scenario. It also validates the HDF5 file format.

This scenario executes the following steps:

  - Creating two test signals
  - Creating two test images
  - Creating a macro
  - Saving current project (h5 file)
  - Removing all objects
  - Opening another project (h5 file)
  - Access current image metadata
  - Checking macro code
"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...
# guitest: show

import os.path as osp

from cdl.env import execenv
from cdl.param import ClipParam
from cdl.tests import cdltest_app_context
from cdl.tests.data import create_noisygauss_image, create_paracetamol_signal
from cdl.tests.scenarios import scenario_mac_app_test
from cdl.utils.strings import save_html_diff
from cdl.utils.tests import CDLTemporaryDirectory


def test_scenario_h5():
    """Example of high-level test scenario with HDF5 file"""
    with CDLTemporaryDirectory() as tmpdir:
        with cdltest_app_context(console=False) as win:
            # === Creating two test signals
            panel = win.signalpanel
            sig1 = create_paracetamol_signal()
            panel.add_object(sig1)
            panel.processor.compute_derivative()
            # === Creating two test images
            panel = win.imagepanel
            ima1 = create_noisygauss_image(add_annotations=True)
            panel.add_object(ima1)
            param = ClipParam.create(upper=ima1.data.mean())
            panel.processor.compute_clip(param)
            # === Creating a macro
            scode = scenario_mac_app_test.add_macro_sample(win, 0).get_code()
            scenario_mac_app_test.add_macro_sample(win, 1)
            # === Saving project
            fname = osp.join(tmpdir, "test.h5")
            win.save_to_h5_file(fname)
            # === Removing all objects
            for panel in win.panels:
                panel.remove_all_objects()
            # === Reopening previously saved project
            win.open_h5_files([fname], import_all=True, reset_all=True)
            # === Accessing object metadata
            obj = win.imagepanel.objmodel.get_groups()[0][0]
            execenv.print(f"Image '{obj.title}':")
            for key, value in obj.metadata.items():
                execenv.print(f'  metadata["{key}"] = {value}')
            # === Checking macro code
            lcode = win.macropanel.get_macro(1).get_code()
            if lcode != scode:
                save_html_diff(scode, lcode, "Saved code", "Loaded code", "macro0.html")
                raise AssertionError("Macro code is not the same as saved")


if __name__ == "__main__":
    test_scenario_h5()



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
.. Data Type Conversion Algorithms (see parent package :mod:`cdl.algorithms`)
"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...

from __future__ import annotations

import numpy as np


def is_integer_dtype(dtype: np.dtype) -> bool:
    """Return True if data type is an integer type

    Args:
        dtype: Data type to check

    Returns:
        True if data type is an integer type
    """
    return issubclass(np.dtype(dtype).type, np.integer)


def is_complex_dtype(dtype: np.dtype) -> bool:
    """Return True if data type is a complex type

    Args:
        dtype: Data type to check

    Returns:
        True if data type is a complex type
    """
    return issubclass(np.dtype(dtype).type, complex)


def clip_astype(data: np.ndarray, dtype: np.dtype) -> np.ndarray:
    """Convert array to a new data type, after having clipped values to the new
    data type's range if it is an integer type.
    If data type is not integer, this is equivalent to ``data.astype(dtype)``.

    Args:
        data: Array to convert
        dtype: Data type to convert to

    Returns:
        Array converted to new data type
    """
    if is_integer_dtype(dtype):
        return np.clip(data, np.iinfo(dtype).min, np.iinfo(dtype).max).astype(dtype)
    return data.astype(dtype)



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
Object model (:mod:`cdl.obj`)
-----------------------------

The :mod:`cdl.obj` module aims at providing all the necessary classes and functions
to create and manipulate DataLab signal and image objects.

Those classes and functions are defined in other modules:
    - :mod:`cdl.core.model.base`
    - :mod:`cdl.core.model.image`
    - :mod:`cdl.core.model.signal`
    - :mod:`cdl.core.io`

The :mod:`cdl.obj` module is thus a convenient way to import all the objects at once.
As a matter of fact, the following import statement is equivalent to the previous one:

.. code-block:: python

    # Original import statement
    from cdl.core.model.signal import SignalObj
    from cdl.core.model.image import ImageObj

    # Equivalent import statement
    from cdl.obj import SignalObj, ImageObj

Common objects
^^^^^^^^^^^^^^

.. autoclass:: cdl.obj.ResultProperties
    :members:
.. autoclass:: cdl.obj.ResultShape
    :members:
.. autoclass:: cdl.obj.ShapeTypes
    :members:
.. autoclass:: cdl.obj.UniformRandomParam
.. autoclass:: cdl.obj.NormalRandomParam
.. autoclass:: cdl.obj.BaseProcParam

Signal model
^^^^^^^^^^^^

.. autodataset:: cdl.obj.SignalObj
    :members:
    :inherited-members:
.. autofunction:: cdl.obj.read_signal
.. autofunction:: cdl.obj.read_signals
.. autofunction:: cdl.obj.create_signal_roi
.. autofunction:: cdl.obj.create_signal
.. autofunction:: cdl.obj.create_signal_from_param
.. autofunction:: cdl.obj.new_signal_param
.. autoclass:: cdl.obj.SignalTypes
.. autodataset:: cdl.obj.NewSignalParam
.. autodataset:: cdl.obj.GaussLorentzVoigtParam
.. autodataset:: cdl.obj.StepParam
.. autodataset:: cdl.obj.PeriodicParam
.. autodataset:: cdl.obj.ROI1DParam
.. autoclass:: cdl.obj.SignalROI

Image model
^^^^^^^^^^^

.. autodataset:: cdl.obj.ImageObj
    :members:
    :inherited-members:
.. autofunction:: cdl.obj.read_image
.. autofunction:: cdl.obj.read_images
.. autofunction:: cdl.obj.create_image_roi
.. autofunction:: cdl.obj.create_image
.. autofunction:: cdl.obj.create_image_from_param
.. autofunction:: cdl.obj.new_image_param
.. autoclass:: cdl.obj.ImageTypes
.. autodataset:: cdl.obj.NewImageParam
.. autodataset:: cdl.obj.Gauss2DParam
.. autodataset:: cdl.obj.ROI2DParam
.. autoclass:: cdl.obj.ImageROI
.. autoclass:: cdl.obj.ImageDatatypes
"""

# pylint:disable=unused-import
# flake8: noqa

from cdl.core.io import read_image, read_images, read_signal, read_signals
from cdl.core.model.base import (
    BaseProcParam,
    NormalRandomParam,
    ResultShape,
    ResultProperties,
    ShapeTypes,
    UniformRandomParam,
)
from cdl.core.model.image import (
    Gauss2DParam,
    ImageDatatypes,
    ImageObj,
    ImageROI,
    ImageTypes,
    NewImageParam,
    ROI2DParam,
    create_image_roi,
    create_image,
    create_image_from_param,
    new_image_param,
)
from cdl.core.model.signal import (
    GaussLorentzVoigtParam,
    NewSignalParam,
    PeriodicParam,
    SignalObj,
    SignalTypes,
    ROI1DParam,
    create_signal_roi,
    StepParam,
    create_signal,
    create_signal_from_param,
    new_signal_param,
    SignalROI,
)
